/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Bootstrap code (restartless add-on)
 *
 * Note:
 * Toolbar button add code, partially taken from Dmitry Gutov's addon "Replace Bookmark"
 * see: https://addons.mozilla.org/fr/firefox/addon/bmreplace/
 *
 * ************************************************************************
 */

"use strict";

Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/devtools/Console.jsm");
Components.utils.import("resource://gre/modules/FileUtils.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");
Components.utils.import("resource://gre/modules/Timer.jsm");

var addonData  = null;
var currReason = null;
var self       = this;
var mDebug     = true;

function loadLib(data) {
  function include(aPath) {
    var path = addonData.resourceURI.spec + "lib/" + aPath + ".js";
    Services.scriptloader.loadSubScript(path, self);
  }

  if (addonData == null) {
    addonData = data;
    Services.scriptloader.loadSubScript(data.resourceURI.spec + "chrome/content/common.js", self);
    Services.scriptloader.loadSubScript(data.resourceURI.spec + "chrome/content/ck.private.js", self);
    mDebug = DEBUG;

    include("l10n");
    include("ui");
    include("preferences");
    include("main");
  } else {
    addonData = data;
  }
}

function logMessage(msg) {
  if (!console) {
    return;
  }
  if (mDebug && addonData != null) {
    console.log("[CookieKeeper " + addonData.version + "] " + msg);
  } else {
    console.log("[CookieKeeper] " + msg);
  }
}

function cmpVersion(a, b) {
  let va = a.split(".");
  let vb = b.split(".");
  let n = va.length < vb.length ? va.length : vb.length; // #define max(a,b) ....
  let r = 0;
  for (let i=0; i<n && r==0; i++) {
    let ia = parseInt(va[i], 10);
    let ib = parseInt(vb[i], 10);
    r = ia - ib;
  }

  return r;
}

function reasonToString(reason) {
  switch (reason) {
  case APP_STARTUP:
    return "APP_STARTUP";
  case ADDON_ENABLE:
    return "ADDON_ENABLE";
  case ADDON_INSTALL:
    return "ADDON_INSTALL";
  case ADDON_UPGRADE:
    return "ADDON_UPGRADE";
  case ADDON_DOWNGRADE:
    return "ADDON_DOWNGRADE";
  case APP_SHUTDOWN:
    return "APP_SHUTDOWN";
  case ADDON_DISABLE:
    return "ADDON_DISABLE";
  case ADDON_UNINSTALL:
    return "ADDON_UNINSTALL";
  }
  return "UNKNOWN (" + reason + ")";
}

function startup(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  APP_STARTUP
  /// &#10;  ADDON_ENABLE
  /// &#10;  ADDON_INSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  const PREF_DELAY = "extensions.cookiekeeper.startup_delay";

  function doStart() {
    try {
      loadLib(data);
      log.d(reasonToString(reason) + ": startup");
      ckMain.init();
    } catch (ex) {
      logMessage(ex);
    }
    self = null;
  }

  currReason = reason;
  let timeout = 0;
  if (Services.prefs.prefHasUserValue(PREF_DELAY)) {
    timeout = Services.prefs.getIntPref(PREF_DELAY);
  }
  if (reason == APP_STARTUP && timeout > 0) {
    setTimeout(doStart, timeout);
  } else {
    doStart();
  }
}

function shutdown(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  APP_SHUTDOWN
  /// &#10;  ADDON_DISABLE
  /// &#10;  ADDON_UNINSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  addonData = data;
  currReason = reason;

  try {
    log.d(reasonToString(reason) + ": shutdown");
    ckMain.uninit();

    if (ADDON_DOWNGRADE === reason) {
      ckPrefs.downgradePreferences();
    }

    if (reason != APP_SHUTDOWN) {
      // http://maglione-k.users.sourceforge.net/bootstrapped.xhtml
      let stringBundleService = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
      stringBundleService.flushBundles();
      // Do a GC when the add-on is disabled, to clear the bootstrap.js code from the memory
      // The bootstrap.js code may still remain resisend a few seconds after the add-on is disabled
      Components.utils.schedulePreciseGC(function() {});
    }

  } catch (ex) {
    logMessage(ex);
  }
}

function install(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  ADDON_INSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  currReason = reason;
  if (reason == ADDON_UPGRADE && cmpVersion(data.oldVersion, "1.7.4") < 0) {
    // Previous version of CookieKeeper (before 1.7.4) do not flushed the string bundle cache
    // on shutdown.
    let Cc = Components.classes;
    let Ci = Components.interfaces;
    let stringBundleService = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService);
    stringBundleService.flushBundles();
  }

  try {
    loadLib(data);
    log.d(reasonToString(reason) + ": install on " + APP_MOZ_CONST[Services.appinfo.ID].NAME);
    ckPrefs.upgradePreferences(reason);
  } catch (ex) {
    logMessage(ex);
  }
}

function uninstall(data, reason) {
  /// <summary>
  /// Bootstrap data structure @see https://developer.mozilla.org/en-US/docs/Extensions/Bootstrapped_extensions#Bootstrap_data
  /// &#10;  string id
  /// &#10;  string version
  /// &#10;  nsIFile installPath
  /// &#10;  nsIURI resourceURI
  /// &#10;
  /// Reason types:
  /// &#10;  ADDON_UNINSTALL
  /// &#10;  ADDON_UPGRADE
  /// &#10;  ADDON_DOWNGRADE
  /// </summary>

  addonData = data;
  try {
    logMessage(reasonToString(reason) + ": uninstall");
    if (ADDON_UNINSTALL === reason) {
      let Cc = Components.classes;
      let Ci = Components.interfaces;
      logMessage("Removing preferences");
      Services.prefs.deleteBranch("extensions.cookiekeeper.");
      let jsonFile = FileUtils.getFile("ProfD", ["cookiekeeper.json"], false);
      if (jsonFile.exists()) {
        logMessage("Delete cookiekeeper.json");
        jsonFile.remove(false);
      }

      // Removing cookiekeeper's localstore.rdf datas
      let localStore = Cc["@mozilla.org/rdf/datasource;1?name=local-store"].getService(Ci.nsIRDFDataSource);
      let enumRes = localStore.GetAllResources();
      let ckRes = [];
      let items = [];

      // Get CookieKeeper resources:
      while (enumRes.hasMoreElements()) {
        let res = enumRes.getNext();
        if(res instanceof Ci.nsIRDFResource) {
          res.QueryInterface(Ci.nsIRDFResource);
          if (res.Value.indexOf("chrome://cookiekeeper/") > -1) {
            ckRes.push(res);
          }
        }
      }

      // For earch resources, get the properties and targets
      for (let i=0; i<ckRes.length; i++) {
        let res = ckRes[i];
        let arcEnum = localStore.ArcLabelsOut(res);
        while (arcEnum.hasMoreElements()) {
          let arc = arcEnum.getNext();
          if (arc instanceof Ci.nsIRDFNode) {
            arc.QueryInterface(Ci.nsIRDFNode);
          }
          if (arc instanceof Ci.nsIRDFResource) {
            arc.QueryInterface(Ci.nsIRDFResource);
          }

          let targetEnum = localStore.GetTargets(res, arc, true);
          while (targetEnum.hasMoreElements()) {
            let target = targetEnum.getNext();
            if (target instanceof Ci.nsIRDFNode) {
              items.push({source: res, property: arc, target: target});
            }
          }
        }
      }
      // Remove from localstore.rdf
      for (let i=0; i<items.length; i++) {
        let item = items[i];
        localStore.Unassert(item.source, item.property, item.target);
      }
    }
  } catch (ex) {
    logMessage(ex.message);
    logMessage(ex.stack);
  }
}
