/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Cookies Storage
 *
 * ************************************************************************
 */

"use strict";

var gCookieStorage = {

  makeCookieObject: function (aCookie) {
    function cookieEquals(aItem) {
      return (typeof(aItem) == "object") &&
        aItem.type  == COOKIE_TYPE_STANDARD &&
        aItem.host  == this.host  &&
        aItem.name  == this.name  &&
        aItem.path  == this.path;
    }
    function getHost() {
      return this.host;
    }
    function getName() {
      return this.name;
    }

    /**
     * Update a cookie with a transform mozilla cookie info.
     * This is used by the observer, when a cookie content is changed.
     */
    function updateFrom(aCookie) {
      if (aCookie.type == COOKIE_TYPE_STANDARD) {
        this.value        = aCookie.value;
        this.isDomain     = aCookie.isDomain;
        this.isSecure     = aCookie.isSecure;
        this.expires      = aCookie.expires;
        this.creationTime = aCookie.creationTime;
        this.expiry       = aCookie.expiry;
        this.isHttpOnly   = aCookie.isHttpOnly;
        this.isSession    = aCookie.isSession;
        this.lastAccessed = aCookie.lastAccessed;
      }
    }

    // Some expiry value have I think the int64 max val: 0x7FFFFFFFFFFFFFFF
    // If so, set it to DATE_MAX instead
    var strippedHost = gCommon.makeStrippedHost(aCookie.host);
    var c = {
      name        : aCookie.name,
      value       : aCookie.value,
      isDomain    : aCookie.isDomain,
      host        : aCookie.host,
      rawHost     : strippedHost,
      path        : aCookie.path,
      isSecure    : aCookie.isSecure,
      expires     : aCookie.expires*1000,
      creationTime: aCookie.creationTime/1000,
      expiry      : aCookie.expiry > DATE_MAX ? DATE_MAX : aCookie.expiry*1000,
      isHttpOnly  : aCookie.isHttpOnly,
      isSession   : aCookie.isSession,
      lastAccessed: aCookie.lastAccessed/1000,
      isProtected : false,
      level       : 1,
      parentIndex : -1,
      type        : COOKIE_TYPE_STANDARD,
      equals      : cookieEquals,
      updateFrom  : updateFrom,
      getHost     : getHost,
      getName     : getName
    };

    c.isProtected = CookieKeeper.isProtectedItem(c);

    if (typeof(aCookie.appId) != "undefined") {
      log.d("Found appId:", aCookie.appId);
    }
    return c;
  },

  /**
   * @brief Fetch the Cookies Storage datas (async load)
   * @param callbackObject an object which implements the function onStorageLoaded(array)
   */
  fetchCookies: function() {
    var cookies     = [];
    var cookiesEnum = Services.cookies.enumerator;

    // Load cookies
    while (cookiesEnum.hasMoreElements()) {
      let cookie = cookiesEnum.getNext();
      if (cookie instanceof Ci.nsICookie2) {
        // Search if it is a protected cookie
        let cookieObj = gCookieStorage.makeCookieObject(cookie);
        cookies.push(cookieObj);
      }
    }

    gCookieKeeper.onStorageLoaded(cookies);
  },

  _performDeletion: function (deleteItems, doBlock, forceBlock) {
    let sp = Services.prefs;
    let sc = Services.cookies;

    var blockFutureCookies = forceBlock;
    if (sp.prefHasUserValue("network.cookie.blockFutureCookies") && !forceBlock) {
      blockFutureCookies = sp.getBoolPref("network.cookie.blockFutureCookies");
    }

    for (let i=0; i<deleteItems.length; i++) {
      let item = deleteItems[i];
      if (item.type == COOKIE_TYPE_STANDARD) {
        sc.remove(item.host, item.name, item.path, blockFutureCookies);
        if (item.isProtected) {
          CookieKeeper.removeProtectedItem(item);
        }
      }
    }
  },

  performDeletion: function (deleteItems, forceBlock) {
    if (deleteItems.length > 0) {
      this._performDeletion(deleteItems, true, forceBlock);
    }
  },

  performDeletionAll: function(deleteItems) {
    if (deleteItems.length > 0) {
      this._performDeletion(deleteItems, false, false);
    }
  }

};
