/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Dom HTML5 Storage
 *
 * Using sqlite module, since there is no interface for it.
 *
 * ************************************************************************
 */

"use strict";

var gDomStorage = {

  makeDomStorageObject: function(aScope, aKey, aValue, aSecure, aOwner) {
    /// Inner DOM Storage item function
    function itemEquals(aItem) {
      return (typeof(aItem) == "object") &&
        aItem.type  == COOKIE_TYPE_STORAGE &&
        aItem.scope == this.scope &&
        aItem.key   == this.key;
    } // function itemEquals()

    function getHost() {
      let host = this.scope.split(":")[0];
      return gCommon.reverseStr(host);
    }
    function getName() {
      return this.key;
    }

    /// Inner DOM Storage item function
    function isJSON() {
      try {
        JSON.parse(this.value);
        return true;
      } catch (ex) {
        return false;
      }
    } // function isJSON()

    function updateFrom(aItem) {
      if (aItem.type  == COOKIE_TYPE_STORAGE) {
        this.value  = aItem.value;
        this.secure = aItem.secure;
        this.owner  = aItem.owner;
      }
    }

    let scopeArray = aScope.split(":");
    let host  = "";
    let proto = "";
    let port  = "";
    switch (scopeArray.length) {
    case 3:
      port  = scopeArray[2];
      // fall-through
    case 2:
      proto = scopeArray[1];
      // fall-through
    case 1:
      host  = gCommon.reverseStr(scopeArray[0]);
    }

    let item = {
      scope       : aScope,
      host        : host,
      rawHost     : gCommon.makeStrippedHost(host),
      proto       : proto,
      port        : port,
      key         : aKey,
      value       : aValue,
      secure      : (typeof(aSecure) == "number" ? aSecure.toString(10) : ""),
      owner       : (typeof(aOwner)  == "string" ? aOwner               : ""),
      path        : "",
      isDomain    : false,
      isSecure    : (proto == "https"),
      expires     : DATE_MAX,
      creationTime: 1000,
      expiry      : DATE_MAX,
      isHttpOnly  : (proto == "http" || proto == "https"),
      isSession   : false,
      lastAccessed: 1000,
      isProtected : false,
      level       : 1,
      parentIndex : -1,
      type        : COOKIE_TYPE_STORAGE,
      equals      : itemEquals,
      updateFrom  : updateFrom,
      isJSON      : isJSON,
      getHost     : getHost,
      getName     : getName
    };
    item.isProtected = CookieKeeper.isProtectedItem(item);

    return item;
  },

  /**
   * @brief Fetch the DOM Storage datas (async load)
   * @param callbackObject an object which implements the function onStorageLoaded(array)
   */
  fetchDOMStorage: function(callbackObject) {
    var storages = [];
    var connect  = null;

    function onRow(aRow) {
      if (aRow == null) {
        return;
      }

      try {
        let scope  = aRow.getString(0);
        let key    = aRow.getString(1);
        let value  = (aRow.getTypeOfIndex(2) == aRow.VALUE_TYPE_TEXT    ? aRow.getString(2) : null);
        let secure = (aRow.getTypeOfIndex(3) == aRow.VALUE_TYPE_INTEGER ? aRow.getInt32(3).toString(10) : "");
        let owner  = (aRow.getTypeOfIndex(4) == aRow.VALUE_TYPE_TEXT    ? aRow.getString(4)             : "");
        if (scope == null || scope.length == 0 || key == null || key.length == 0) {
          return;
        }

        let item = gDomStorage.makeDomStorageObject(scope, key, value, secure, owner);
        storages.push(item);
      } catch (ex) {
        Cu.reportError(CK_MODULE_NAME + " Fetch DOM storage (create item) exception: " + ex);
        Cu.reportError(ex);
        throw ex;
      }
    } // function onRow()

    function onSelectComplete(result) {
      try {
        if (typeof(result) == "boolean") {
          if (result == false) {
            Cu.reportError(CK_MODULE_NAME + " Fetch DOM storage (SELECT) error");
          }
        } else if (typeof(result) != "undefined" && result != null) {
          Cu.reportError(CK_MODULE_NAME + " Fetch DOM storage (SELECT) exception: " + result);
        }
        try {
          gCookieKeeper.onStorageLoaded(storages);
        } catch (ex) {
          Cu.reportError(CK_MODULE_NAME + " Fetch DOM storage (close connection) exception: " + ex);
          Cu.reportError(ex);
        }
      } finally {
        connect.close();
      }
    } // function onSelectComplete()

    function doSelect(conn) {

      connect = conn;
      try {
        let res = connect.execute("SELECT scope, key, value, secure, owner FROM " + DOM_DB_TABLE, null, onRow);
        res.then(onSelectComplete, onSelectComplete);
      } catch (ex) {
        Cu.reportError(CK_MODULE_NAME + " Fetch DOM storage (connection) exception:" + ex);
        Cu.reportError(ex);
        gCookieKeeper.onStorageLoaded([]);
        throw ex;
      }
    } // function doSelect()

    let connection = Sqlite.openConnection({path: DOM_DB_NAME});
    connection.then(doSelect, Cu.reportError);
  },

  /**
   * Delete dom storage items
   * @param aItems: An array of DOM Storage items to delete.
   * @return true if the delete is launch (aItem is not null and contains data)
   */
  performDeletion: function (aItems) {
    var db    = null;
    var error = null;

    function doTransaction() {
      let nItems         = aItems.length;
      for (let i=0; i<nItems; i++) {
        let item = aItems[i];
        if (item.type  == COOKIE_TYPE_STORAGE) { // safety check
          yield db.execute("DELETE FROM " + DOM_DB_TABLE + " WHERE scope=? and key=?", [item.scope, item.key], null);
        }
      }
    }

    function doDelete() {
      try {
        db = yield Sqlite.openConnection({path: DOM_DB_NAME});
        yield db.executeTransaction(doTransaction);
      } catch (ex) {
        log.e("Delete DOM Storage:" + ex.message, ex.stack);
        error = ex;
      } finally {
        if (db != null) {
          db.close();
          db = null;
        }
        // Item are deleted, notify observers
        let subject = {
          error  : error,
          deleted: aItems
        }
        Services.obs.notifyObservers({ wrappedJSObject: subject }, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_DELETE);
      }
    }

    if (aItems == null || aItems.length == 0) {
      return false;
    }
    Task.spawn(doDelete);
    return true;
  },

  /**
   * @brief Insert items in the webappsstore.sqlite database.
   * If an item already exists, it will be updated.
   * @param aItems An array of dom storage items to insert.
   * @return true if the insert is launch (aItem is not null and contains data)
   * @note
   * Each item must be in this form:
   * let item = {scope:"ascope", key:"akey", value:"avalue"};
   * Note that the scope must be in db format: the domain name must be reversed
   * (eg: gro.allizom.snodda.:https:443 for addons.mozilla.org)
   */
  performInsertion: function (aItems) {
    var db    = null;
    var error = null;
    var inserted = [];

    function doTransaction() {

      let nItems = aItems.length;
      for (let i=0; i<nItems; i++) {
        let item = aItems[i];
        yield db.execute("INSERT OR REPLACE INTO " + DOM_DB_TABLE + " (scope,key,value) VALUES (?,?,?)", [item.scope, item.key, item.value]);
      }
    }
    function doInsert() {
      try {
        db = yield Sqlite.openConnection({path: DOM_DB_NAME});
        yield db.executeTransaction(doTransaction);

        inserted = aItems;
      } catch (ex) {
        log.e("Insert DOM Storage:" + ex.message, ex.stack);
        error = ex;
      } finally {
        if (db != null) {
          db.close();
          db = null;
        }
        // Notify the update is done
        let subject = {
          error : error,
          insert: inserted
        }
        Services.obs.notifyObservers({ wrappedJSObject: subject }, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_INSERT);
      }
    }

    if (aItems == null || aItems.length == 0) {
      return false;
    }
    Task.spawn(doInsert);
    return true;
  }
};
