/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Main window functions
 *
 * ************************************************************************
 */

"use strict";

const BUNDLE_PF_ALL           = "cookiesAll";
const BUNDLE_PF_ALL_PROT      = "cookiesAllProtected";
const BUNDLE_PF_ALL_UNPROT    = "cookiesAllUnprotected";
const BUNDLE_PF_FILTER        = "cookiesFiltered";
const BUNDLE_PF_FILTER_PROT   = "cookiesFilteredProtected";
const BUNDLE_PF_FILTER_UNPROT = "cookiesFilteredUnprotected";

/**
 * Cookie Keeper Window management:
 */
var gCookieKeeper = {
  // TODO: when it is possible to access dom storage
  // See bugs:
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=343163
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=506692
  // - https://bugzilla.mozilla.org/show_bug.cgi?id=527667
  // _sDomStorage          : Cc["@mozilla.org/dom/localStorage-manager;1"].getService(Ci.nsIDOMStorageManager),

  // UI Variables
  _bundle           : null, // String translations (cookie culler specific)
  _tree             : null, // Cookie XUL tree (cookiesTree)

  // String
  _yes              : "yes",
  _no               : "no",

  _labelProtec      : null, // Used to quickly set the right protect label for the context menu
  _labelRemove      : null, // Used to quickly set the right remove label for the context menu

  _blockSelectEvent : 0,
  _lastSort         : SORT_INFOS.DEFAULT,
  _lastSortAscending: true,
  // Initial load checks:
  _nStorageToLoad   : 0,
  _nStorageLoaded   : 0,
  _nTimeout         : 0,

  init: function() {
    try {
      this._bundle   = document.getElementById("bundleCookieKeeper");
      this._tree     = document.getElementById("cookiesTree");
      this._yes      = this._bundle.getString("yes");
      this._no       = this._bundle.getString("no");

      this._tree.treeBoxObject.view = gCookiesTreeView;
      gCookiesTreeView.init(this._bundle);

      gTestMozPrefs.doTest();
      gCookiesObserver.init();

      this.reload(false);
    } catch (ex) {
      Cu.reportError(ex);
    }
  },

  uninit: function() {
    try {
      this.blockSelectEvent();
      gCookiesObserver.uninit();
      gCookiesTreeView.reset();

      this._bundle      = null;
      this._tree        = null;
    } catch (ex) {
      Cu.reportError(ex);
    }
  },

  blockSelectEvent: function() {
    this._blockSelectEvent++;
  },


  unblockSelectEvent: function() {
    this._blockSelectEvent--;
  },

  clear: function() {
    document.getElementById(this._lastSort.column).removeAttribute("sortDirection");

    this._blockSelectEvent  = 1;
    this._lastSort          = SORT_INFOS.DEFAULT;
    this._lastSortAscending = true;
    this._nStorageToLoad    = 0;
    this._nStorageLoaded    = 0;
    this._nTimeout          = 0;

    document.getElementById(SORT_INFOS.DEFAULT.column).setAttribute("sortDirection", "ascending");
    document.getElementById("labelProtectionFilter").value = this._bundle.getString(BUNDLE_PF_ALL);

    gCookiesTreeView.reset();

    this._updateButtons();
    this._updateItemData();
    this._blockSelectEvent = 0;
  },

  reload: function(doSaveState) {
    let lastSort    = this._lastSort;
    let lastSortAsc = this._lastSortAscending;
    if (doSaveState) {
      gCookiesTreeView.saveState();
    }

    // Block the added cookie event, to prevent adding cookies 2x,
    // if new cookies are created during the loading.
    // This can happend if the cookies keeper window is display
    // and we select "Delete all cookies" from menu or toolbar button.
    gCookiesObserver.blockCookiesEvents     = true;
    gCookiesObserver.blockProtectionsEvents = true;
    this.clear();
    gCookieKeeper.blockSelectEvent();

    // Launch async tasks to load cookies and DOM storage
    try {
      gCookieStorage.fetchCookies();
      this._nStorageToLoad++;
      gDomStorage.fetchDOMStorage();
      this._nStorageToLoad++;
    } catch (ex) {
      log.e("Error fetch:", ex);
    }

    if (doSaveState) {
      this._lastSort          = lastSort;
      this._lastSortAscending = lastSortAsc;
    }

    // Set a timeout to display the result
    // Using a inner function in the first argument,
    // to silence a warning in the add-ons validator.
    window.setTimeout(function() { gCookieKeeper.onLoadTimeout(doSaveState); }, 100);
  },

  onLoadTimeout: function(doRestoreState) {
    function onLoaded() {
      try {
        gCookiesTreeView.refreshView(doRestoreState);
        gCookieKeeper.unblockSelectEvent();
        if (doRestoreState) {
          let lastSort = gCookieKeeper._lastSort;
          gCookieKeeper._lastSort = SORT_INFOS.DEFAULT;
          gCookieKeeper.onSort(null, lastSort, gCookieKeeper._lastSortAscending);
          // Hack to be sure the column sortDirection attribute will be reflected in the view:
          gCookieKeeper.focusFilterBox();
          gCookieKeeper._tree.focus();
        } else {
          gCookieKeeper.onItemSelected(null);
          gCookieKeeper.focusFilterBox();
        }
      } finally {
        gCookiesObserver.blockCookiesEvents     = false;
        gCookiesObserver.blockProtectionsEvents = false;
        gCookieKeeper._nStorageToLoad = 0;
        gCookieKeeper._nStorageLoaded = 0;
        gCookieKeeper._nTimeout       = 0;
      }
    }

    if (gCookieKeeper._nStorageLoaded == gCookieKeeper._nStorageToLoad || gCookieKeeper._nTimeout >= 10) {
      onLoaded();
    } else {
      gCookieKeeper._nTimeout++;
      window.setTimeout(function() { gCookieKeeper.onLoadTimeout(doRestoreState); }, 100);
    }
  },

  onStorageLoaded: function(aItemsArray) {
    // Callback function used when items are loaded
    this._nStorageLoaded++;
    gCookiesTreeView.addItems(aItemsArray);
  },

  _updateItemData: function () {
    var seln       = gCookiesTreeView.selection;
    var disabled   = true;
    var properties = null;
    var item       = seln.currentIndex == -1 ? null : gCookiesTreeView.getItemAtIndex(seln.currentIndex);
    var deck       = document.getElementById("itemDeck");

    var imgProtWarnSession = document.getElementById("imgCookieIsProtectedWarnSession");
    imgProtWarnSession.hidden = true;

    if (item == null || item.type == COOKIE_TYPE_CONTAINER || seln.count > 1) {
      let noneSelected = this._bundle.getString("noCookieSelected");

      properties = {
        tbCookieHost: noneSelected,
        tbCookieName: noneSelected,
        tbCookieValue: noneSelected,
        tbCookiePath: noneSelected,
        tbCookieIsSecure: noneSelected,
        tbCookieExpires: noneSelected,
        tbCookieIsProtected: noneSelected
      };
      deck.selectedPanel = document.getElementById("cookieInfoGrid");;

    } else if (item.type == COOKIE_TYPE_STORAGE) {
      let secure = item.isSecure ?
        this._bundle.getString("forSecureOnly") :
        this._bundle.getString("forAnyConnection");

      let value;
      if (item.value.length > 1024) {
        value = item.value.substring(0, 1024);
        document.getElementById("btnShowValue").hidden = false;
      } else {
        value = item.value;
        // If the first char is '{' it may be a json value
        // Propose then to display the "more box" which permit to format JSON objects
        document.getElementById("btnShowValue").hidden = !item.isJSON();
      }
      properties = {
        tbDomHost       : item.host,
        tbDomPort       : item.port,
        tbDomProtocol   : item.proto,
        tbDomKey        : item.key,
        tbDomValue      : value,
        tbDomIsProtected: item.isProtected ? this._yes : this._no
      };

      deck.selectedPanel = document.getElementById("domStorageInfoGrid");
      disabled = false;

    } else { // Cookie
      let expires = item.isSession ?
        this._bundle.getString("expireAtEndOfSession") :
        gCommon.formatDateString(item.expires, null); // this._bundle.getString("expireAtEndOfSession")
      let secure = item.isSecure ?
        this._bundle.getString("forSecureOnly") :
        this._bundle.getString("forAnyConnection");
      let domain = item.isDomain ?
        this._bundle.getString("domainColon") :
        this._bundle.getString("hostColon");

      properties = {
        tbCookieHost       : item.host,
        labelCookieDomain  : domain,
        tbCookieName       : item.name,
        tbCookieValue      : item.value,
        tbCookiePath       : item.path,
        tbCookieExpires    : expires,
        tbCookieIsSecure   : secure,
        tbCookieIsProtected: item.isProtected ? this._yes : this._no
      };
      deck.selectedPanel = document.getElementById("cookieInfoGrid");
      disabled = false;

      imgProtWarnSession.hidden = !(item.isProtected && item.isSession);
    }

    for (let property in properties) {
      let elem      = document.getElementById(property);
      elem.value    = properties[property];
      elem.disabled = disabled;
    }
  },

  _updateButtonsProtect: function(aItemsArray, btProtec, btUnProt) {
    // log.d("gCookieKeeper._updateButtonsProtect()");
    // Check if we have to protect or unprotect
    // if mixed content, choose protection
    const UNDEF_PROTECT = 0;
    const ALL_PROTEC    = 1;
    const ALL_UNPROTECT = 2;
    const MIXED_PROTEC  = 3;
    var   protection    = UNDEF_PROTECT;

    for (let i=0; i<aItemsArray.length && protection != MIXED_PROTEC; i++) {
      var item = aItemsArray[i];

      switch (protection) {
      case UNDEF_PROTECT:
        protection = (item.isProtected ? ALL_PROTEC : ALL_UNPROTECT);
        break;
      case ALL_PROTEC:
        protection = (item.isProtected ? ALL_PROTEC : MIXED_PROTEC);
        break;
      case ALL_UNPROTECT:
        protection = (item.isProtected ? MIXED_PROTEC : ALL_UNPROTECT);
        break;
      }
    }

    switch (protection) {
    case ALL_PROTEC:
      // all cookies protected, we can unprotect them at once
      btProtec.disabled = true;
      btUnProt.disabled = false;
      btUnProt.parentNode.selectedPanel = btUnProt;
      this._labelProtec = btUnProt.label;
      break;
    case ALL_UNPROTECT:
    case MIXED_PROTEC:
      btProtec.disabled = false;
      btUnProt.disabled = true;
      btProtec.parentNode.selectedPanel = btProtec;
      this._labelProtec = btProtec.label;
      break;
    case ALL_DOM_STORE:
      btProtec.disabled = true;
      btUnProt.disabled = true;
      btProtec.parentNode.selectedPanel = btProtec;
      this._labelProtec = btProtec.label;
    }
  },

  _updateButtons: function () {
    // log.d("gCookieKeeper._updateButtons()");
    var disabled = false;

    var tbFilter = document.getElementById("textBoxSearch"); // filter text box
    var btRemAll = document.getElementById("removeAllCookies");
    var btRemove = document.getElementById("removeCookie");
    var btRemMul = document.getElementById("removeCookies");
    var btProtec = document.getElementById("protectCookie");
    var btUnProt = document.getElementById("unprotectCookie");
    var btEdit   = document.getElementById("editButton");

    var seln = gCookiesTreeView.selection;

    if (gCookiesTreeView.viewMode == VIEW_MODE.FLAT) {
      disabled = gCookiesTreeView.rowCount == 0;
      // Never disable the text box filter when the filter is set
      if (gCookiesTreeView.searchValue.length > 0) {
        tbFilter.disabled = false;
      } else {
        tbFilter.disabled = disabled;
      }
    } else {
      disabled = (Object.keys(gCookiesTreeView._dHosts).length == 0);
      tbFilter.disabled = disabled;
      // debug:
      if (Object.keys(gCookiesTreeView._dHosts).length > gCookiesTreeView.rowCount) {
        log.e("_updateButtons: row count error", Object.keys(gCookiesTreeView._dHosts).length, gCookiesTreeView.rowCount, gCookiesTreeView);
      }
    }

    btRemAll.disabled = disabled;
    btEdit.disabled   = true;

    if (disabled || seln.currentIndex == -1) {
      // No cookie is displayed
      // or no item is selected
      // disable all
      btRemove.disabled = true;
      btRemMul.disabled = true;
      btProtec.disabled = true;
      btUnProt.disabled = true;

      btRemove.parentNode.selectedPanel = btRemove;
      btProtec.parentNode.selectedPanel = btProtec;
      this._labelRemove = btRemove.label;
      this._labelProtec = btProtec.label;

    } else {
      // There is some cookies displayed, and at least one selected
      if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
        // In tree view there is only one selected item max
        var item = gCookiesTreeView.getItemAtIndex(seln.currentIndex);
        // item can't be null here, or there is a bug somewhere
        if (item == null) {
          log.e("_updateButtons: Item is null at selection index " + seln.currentIndex);
          return;
        }
        if (item.type == COOKIE_TYPE_CONTAINER) {
          if (item.items == null || item.items.length == 0) {
            // debug
            log.e("_updateButtons: no cookie in container!");
          }

          if (item.items.length > 1) {
            btRemove.disabled = true;
            btRemMul.disabled = false;
            btRemMul.parentNode.selectedPanel = btRemMul;
            this._updateButtonsProtect(item.items, btProtec, btUnProt);
            this._labelRemove = btRemMul.label;
          } else {
            // One cookie in container
            btRemove.disabled = false;
            btRemMul.disabled = true;
            btRemove.parentNode.selectedPanel = btRemove;
            this._labelRemove = btRemove.label;
            this._updateButtonsProtect(item.items, btProtec, btUnProt);
          }

        } else {
          // item is a cookie
          btRemove.disabled = false;
          btRemMul.disabled = true;
          btEdit.disabled   = false;
          btRemove.parentNode.selectedPanel = btRemove;
          this._labelRemove = btRemove.label;
          this._updateButtonsProtect([item], btProtec, btUnProt);
        }

      } else {
        // flat & filter view
        var rangeCount          = seln.getRangeCount();
        var selectedCookieCount = 0;
        var items               = [];

        // Get all selected cookies
        for (let i = 0; i < rangeCount; i++) {
          var min = {}; var max = {};
          seln.getRangeAt(i, min, max);
          for (let j = min.value; j <= max.value; j++) {
            let item = gCookiesTreeView.getItemAtIndex(j);
            if (item == null) {
              log.e("item at index " + j + " is null");
              continue;
            }
            selectedCookieCount++;
            items.push(item);
          }
        }

        if (selectedCookieCount == 0) {
          log.e("_updateButtons: no item selected");
        } else if (selectedCookieCount == 1) {
          btRemove.disabled = false;
          btRemMul.disabled = true;
          btEdit.disabled   = false;
          btRemove.parentNode.selectedPanel = btRemove;
          this._labelRemove = btRemove.label;

          this._updateButtonsProtect(items, btProtec, btUnProt);
        } else {
          // more than 1 cookie selected
          btRemove.disabled = true;
          btRemMul.disabled = false;
          btRemMul.parentNode.selectedPanel = btRemMul;
          this._labelRemove = btRemMul.label;
          this._updateButtonsProtect(items, btProtec, btUnProt);
        }
      }
    }
  },

  onItemSelected: function (event) {
    // log.d("gCookieKeeper.onItemSelected(): ", this._blockSelectEvent);
    if (this._blockSelectEvent == 0) {
      this._updateButtons();
      this._updateItemData();
    } else if (this._blockSelectEvent < 0) {
      log.e("this._blockSelectEvent < 0: ", this._blockSelectEvent)
    }
  },

  _askDelCookie: function(aCount) {
    var doNotAsk = false;
    var doDel    = true;

    if (aCount == 0) {
      return false;
    }

    if (Services.prefs.prefHasUserValue(PREF_NO_ASK_AGAIN_DEL_PROT)) {
      doNotAsk = Services.prefs.getBoolPref(PREF_NO_ASK_AGAIN_DEL_PROT);
    }

    if (!doNotAsk) {
      var prompts     = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var check       = {value: false};
      var flags       = prompts.BUTTON_TITLE_YES    * prompts.BUTTON_POS_0 +
                        prompts.BUTTON_TITLE_NO     * prompts.BUTTON_POS_1;
      var sDelProt    = this._bundle.getString(aCount > 1 ? "del_protects" : "del_protect");
      var sNoAskAgain = this._bundle.getString("no_ask_again");
      var button      = prompts.confirmEx(window, "CookieKeeper", sDelProt, flags, "", "", "", sNoAskAgain, check);

      if (button == 0) {
        if (check.value) {
          // setup the preferences
          Services.prefs.setBoolPref(PREF_NO_ASK_AGAIN_DEL_PROT, true);
        }
      } else {
        doDel = false;
      }
    }

    return doDel;
  },

  onDeleteCookie: function (doBlock) {
    var cookies      = [];
    var storage      = [];
    var nProtected   = 0;
    var delProtected = true;
    var seln         = gCookiesTreeView.selection;

    if (seln.count < 1) {
      return;
    }

    if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
      let ci = seln.currentIndex;
      let item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

      if (item == null) {
        log.e("No item at index " + ci);
        return;
      }

      if (item.type == COOKIE_TYPE_CONTAINER) {
        for (let i=0; i<item.items.length; i++) {
          let cookie = item.items[i];
          nProtected += cookie.isProtected ? 1 : 0;
          switch (cookie.type) {
          case COOKIE_TYPE_STANDARD:
            cookies.push(cookie);
            break;
          case COOKIE_TYPE_STORAGE:
            storage.push(cookie);
            break;
          default:
            log.e("Unknwow cookie type " + cookie.type);
            break;
          }
        }
      } else {
        nProtected = item.isProtected ? 1 : 0;
        switch (item.type) {
        case COOKIE_TYPE_STANDARD:
          cookies.push(item);
          break;
        case COOKIE_TYPE_STORAGE:
          storage.push(item);
          break;
        default:
          log.e("Unknwow cookie type " + cookie.type);
          break;
        }
      }
    } else {
      // VIEW_MODE.FLAT
      let rangeCount = seln.getRangeCount();

      for (let i = rangeCount - 1; i >= 0; i--) {
        let min = {}; var max = {};
        seln.getRangeAt(i, min, max);

        for (let j = min.value; j <= max.value; j++) {
          let item = gCookiesTreeView.getItemAtIndex(j);
          if (typeof(item) == "undefined" || item == null) {
            log.e("No item at index " + j);
            return;
          }

          nProtected += item.isProtected ? 1 : 0;
          switch (item.type) {
          case COOKIE_TYPE_STANDARD:
            cookies.push(item);
            break;
          case COOKIE_TYPE_STORAGE:
            storage.push(item);
            break;
          default:
            log.e("Unknwow cookie type " + cookie.type);
            break;
          }
        }
      }
    }

    // common part:
    if (nProtected > 0) {
      delProtected = this._askDelCookie(nProtected);
    }

    if (delProtected) {
      gCookieStorage.performDeletion(cookies, doBlock);
      gDomStorage.performDeletion(storage);
    } else {
      let cookies2 = [];
      let storage2 = [];
      for (let i=0; i<cookies.length; i++) {
        let cookie = cookies[i];
        if (!cookie.isProtected) {
          cookies2.push(cookie);
        }
      }
      for (let i=0; i<storage.length; i++) {
        let item = storage[i];
        if (!item.isProtected) {
          storage2.push(item);
        }
      }
      gCookieStorage.performDeletion(cookies2, doBlock);
      gDomStorage.performDeletion(storage2);
    }

    this._tree.focus();
  },

  /// Delete all displayed and unprotected cookies
  onDeleteAllCookies: function () {
    const viewMode  = gCookiesTreeView.viewMode;
    var cookies = [];
    var storage = [];

    if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
      for (let key in gCookiesTreeView._dHosts) {
        var host = gCookiesTreeView._dHosts[key];
        let nc = host.items.length;
        for (let i=0; i<nc; i++) {
          var item = host.items[i];
          if (!item.isProtected) {
            switch (item.type) {
            case COOKIE_TYPE_STANDARD:
              cookies.push(item);
              break;
            case COOKIE_TYPE_STORAGE:
              storage.push(item);
              break;
            default:
              log.e("Unknwow cookie type " + cookie.type);
              break;
            }
          }
        }
      }
    } else {
      let flatSet = gCookiesTreeView._flatSet;
      let count   = flatSet.length;
      for (let i=0; i<count; i++) {
        let item = flatSet[i];
        if (!item.isProtected) {
          switch (item.type) {
          case COOKIE_TYPE_STANDARD:
            cookies.push(item);
            break;
          case COOKIE_TYPE_STORAGE:
            storage.push(item);
            break;
          default:
            log.e("Unknwow cookie type " + cookie.type);
            break;
          }
        }
      }
    }

    if (cookies.length > 0 || storage.length > 0) {
      gCookiesObserver.blockCookiesEvents = true;
      this.blockSelectEvent();
      gCookiesTreeView.saveState();

      if (cookies.length > 0) {
        gCookieStorage.performDeletionAll(cookies);
        // Remove cookies from the gCookiesTreeView
        for (let i=0; i<cookies.length; i++) {
          gCookiesTreeView.deleteItemFromCache(cookies[i]);
        }
      }

      if (storage.length > 0) {
        gDomStorage.performDeletion(storage);
      }
      gCookiesTreeView.refreshView(true);

      gCookiesObserver.blockCookiesEvents = false;
      this.unblockSelectEvent();
      this.onItemSelected(null);
    }
    this._tree.focus();
  },

  onProtectCookie: function(aEvent) {
    const seln = gCookiesTreeView.selection;
    if (seln.count == 0) {
      return;
    }
    const ci = seln.currentIndex;
    if (ci == -1) {
      return;
    }

    // use the button state, faster than check all cookies
    // to know if we have to protect or unprotect
    const doProtect = !document.getElementById("protectCookie").disabled;

    let items   = [];
    let indexes = [];

    // Get items
    if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
      let item = gCookiesTreeView.getItemAtIndex(ci);

      if (item.type == COOKIE_TYPE_CONTAINER) {
        for (let i=0; i<item.items.length; i++) {
          var cookie = item.items[i];
          items.push(cookie);
          indexes.push((ci + i + 1));
        }
      } else {
        items.push(item);
        indexes.push(ci);
      }

    } else {
      // Flat & filter view
      var rangeCount = seln.getRangeCount();

      // For all selected cookies
      for (let i = 0; i < rangeCount; i++) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);
        for (let j = min.value; j <= max.value; j++) {
          var cookie = gCookiesTreeView.getItemAtIndex(j);
          items.push(cookie);
          indexes.push(j);
        }
      }
    }

    // Perform the change + Update UI
    let tbo = gCookiesTreeView._treeBoxObj;
    tbo.beginUpdateBatch();
    gCookiesObserver.blockProtectionEvent = true;
    let len = items.length;
    if (doProtect) {
      let nSession = 0; // Number of protected session cookies
      for (let i=0; i<len; i++) {
        let item = items[i];
        if (!item.isProtected) {
          CookieKeeper.addProtectedItem(item);
          item.isProtected = true;
          if (item.isSession) {
            nSession++;
          }
          tbo.invalidateRow(indexes[i]);
        }
      }

      if (nSession > 0) {
        // Show a warning notification
        let notification    = document.createElementNS(XUL_NS, "notification");
        let notificationBox = document.getElementById("notificationBox");
        notification.setAttribute("id", "notifyProt");
        notification.setAttribute("type", "warning");
        notification.setAttribute("image", "chrome://cookiekeeper/skin/warning16.png");
        notification.setAttribute("label", this._bundle.getString("notification.cookie.protect_session"));
        notificationBox.appendChild(notification);
      }
    } else {
      for (let i=0; i<len; i++) {
        let item = items[i];
        if (item.isProtected) {
          CookieKeeper.removeProtectedItem(item);
          item.isProtected = false;
          tbo.invalidateRow(indexes[i]);
        }
      }
    }

    gCookiesObserver.blockProtectionEvent = false;
    tbo.endUpdateBatch();
    this.onItemSelected(null);
    this._tree.focus();
  },

  onShowEditor: function() {
    var seln = gCookiesTreeView.selection;
    if (seln.count == 0) {
      return;
    }
    var ci = seln.currentIndex;
    var item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);
    if (item == null || item.type == COOKIE_TYPE_CONTAINER) {
      this._tree.focus();
      return;
    }

    if (item.type == COOKIE_TYPE_STORAGE) {
      window.openDialog("chrome://cookiekeeper/content/ck.editor.dom_storage.xul",
        "cookieKeeperEditorDomStorage",
        "chrome,centerscreen,resizable,modal",
        item, true
      );
    } else {
      window.openDialog("chrome://cookiekeeper/content/ck.editor.cookie.xul",
        "cookieKeeperEditorCookie",
        "chrome,centerscreen,resizable,modal",
        item
      );
    }

    this._tree.focus();
  },

  onShowDomViewer: function() {
    var seln = gCookiesTreeView.selection;
    if (seln.count == 0) {
      return;
    }
    var ci   = seln.currentIndex;
    var item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

    if (item == null || item.type == COOKIE_TYPE_CONTAINER) {
      this._tree.focus();
      return;
    }

    window.openDialog("chrome://cookiekeeper/content/ck.editor.dom_storage.xul",
      "cookieKeeperEditorDomStorage",
      "chrome,modal,titlebar",
      item, false
    );
    this._tree.focus();
  },

  onCookieKeyPress: function (aEvent) {
    // log.d("onCookieKeyPress", aEvent);
    switch (aEvent.charCode) {
      case aEvent.DOM_VK_SPACE:
        this.onProtectCookie(aEvent);
        aEvent.stopPropagation();
        return;
      default:
        break;
    }
    switch (aEvent.keyCode) {
      case aEvent.DOM_VK_DELETE:
        this.onDeleteCookie(false);
        aEvent.stopPropagation();
        return;
      default:
        break;
    }
  },

  onWindowKeyPress: function (aEvent) {
    if (aEvent.keyCode == KeyEvent.DOM_VK_ESCAPE) {
      window.close();
    }
  },

  onContextMenu: function (aEvent) {
    // Prevent to show the context menu if there is no selection, or nothing displayed
    // gCookiesTreeView.selection.count may return 1, even if the tree view is empty...
    var showContextMenu = (gCookiesTreeView.rowCount > 0 && gCookiesTreeView.selection.count > 0);
    if (!showContextMenu) {
      // Seems to work better to prevent the popup showing, than return false.
      aEvent.preventDefault();
    }
    return showContextMenu;
  },

  onShowContextCookieMenu: function(aEvent) {
    // The context menu can't show up if there is no selection
    // see onContextMenu() function

    var sp   = Services.prefs;
    var seln = gCookiesTreeView.selection;

    var mpccPerms      = document.getElementById("mpcc-perms");
    var mpccSepPerm    = document.getElementById("mpcc-sep-perm");

    var mpccProtect    = document.getElementById("mpcc-protect");
    var mpccEdit       = document.getElementById("mpcc-edit");
    var mpccDelete     = document.getElementById("mpcc-delete");
    var mpccDelBlock   = document.getElementById("mpcc-delete-and-block");

    var mpccSepCopy    = document.getElementById("mpcc-sep-copy");
    var mpccCopyDomain = document.getElementById("mpcc-copy-domain");
    var mpccCopyName   = document.getElementById("mpcc-copy-name");
    var mpccCopyValue  = document.getElementById("mpcc-copy-value");

    // Update the context menu label/hidden properties:

    mpccProtect.label = this._labelProtec;

    mpccDelBlock.hidden = false;
    if (sp.prefHasUserValue("network.cookie.blockFutureCookies")) {
      // Don't show the delete and block context menu item
      // if we have to do it by default.
      if (sp.getBoolPref("network.cookie.blockFutureCookies")) {
        mpccDelBlock.hidden = true;
      }
    }

    if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
      let ci = seln.currentIndex;
      let item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

      if (item == null) {
        log.e("onShowContextCookieMenu: item is null at selected index " + ci);
        return false;
      }

      mpccPerms.hidden   = false;
      mpccSepPerm.hidden = false;

      if (item.type == COOKIE_TYPE_CONTAINER) {
        mpccEdit.hidden       = true;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = true;
        mpccCopyValue.hidden  = true;

      } else { // a cookie
        mpccEdit.hidden       = false;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = false;
        mpccCopyValue.hidden  = false;
      }

    } else {
      // Flat view
      if (seln.count > 1) {
        mpccPerms.hidden   = true;
        mpccSepPerm.hidden = true;

        mpccEdit.hidden       = true;
        mpccSepCopy.hidden    = true;
        mpccCopyDomain.hidden = true;
        mpccCopyName.hidden   = true;
        mpccCopyValue.hidden  = true;
      } else {
        mpccPerms.hidden   = false;
        mpccSepPerm.hidden = false;

        mpccEdit.hidden       = false;
        mpccSepCopy.hidden    = false;
        mpccCopyDomain.hidden = false;
        mpccCopyName.hidden   = false;
        mpccCopyValue.hidden  = false;
      }
    }

    return true;
  },

  onShowContextCookiePerms: function(aEvent) {
    const sCPerm = Cc["@mozilla.org/cookie/permission;1"].getService(Ci.nsICookiePermission);

    var seln = gCookiesTreeView.selection;
    var ci   = seln.currentIndex;
    var item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

    if (item == null) {
      log.e("onShowContextCookiePerms: Item is null at current selection index " + ci);
      return false;
    }

    var aHost = item.rawHost;

    var mpccDefault = document.getElementById("mpcc-perm-default");
    var mpccAllow   = document.getElementById("mpcc-perm-allow");
    var mpccDeny    = document.getElementById("mpcc-perm-deny");
    var mpccSession = document.getElementById("mpcc-perm-session");
    // var mpccFirst   = document.getElementById("mpcc-perm-first");
    // var mpccThird   = document.getElementById("mpcc-perm-third");

    mpccDefault.setAttribute("checked", false);
    mpccAllow.setAttribute("checked", false);
    mpccDeny.setAttribute("checked", false);
    mpccSession.setAttribute("checked", false);
    // mpccFirst.setAttribute("checked", false);
    // mpccThird.setAttribute("checked", false);

    if (item.type != COOKIE_TYPE_CONTAINER) {
      aHost = item.host;
    }
    aHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;

    var uri  = Services.io.newURI("http://" + aHost, null, null);
    var perm = Services.perms.testPermission(uri, "cookie");

    switch (perm) {
    case sCPerm.ACCESS_DEFAULT:
      mpccDefault.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_ALLOW:
      mpccAllow.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_DENY:
      mpccDeny.setAttribute("checked", true);
      break;
    case sCPerm.ACCESS_SESSION:
      mpccSession.setAttribute("checked", true);
      break;
    // case sCPerm.ACCESS_ALLOW_FIRST_PARTY_ONLY:
    //   mpccFirst.setAttribute("checked", true);
    //   break;
    // case sCPerm.ACCESS_LIMIT_THIRD_PARTY:
    //   mpccThird.setAttribute("checked", true);
    //   break;
    }

    return true;
  },

  onCookieDBLClick: function (aEvent) {
    var seln = gCookiesTreeView.selection;
    if (seln.count == 0) {
      return;
    }
    var ci = seln.currentIndex;
    var item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);
    if (item != null && item.type != COOKIE_TYPE_CONTAINER) {
      this.onProtectCookie(aEvent);
    }
  },

  onContextMenuPermission: function(aPermission) {
    var seln = gCookiesTreeView.selection;
    var ci   = seln.currentIndex;
    var item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

    if (item == null) {
      // TODO Error dialog
      return;
    }

    var aHost = item.rawHost;

    if (item.type != COOKIE_TYPE_CONTAINER) {
      aHost = item.host;
    }
    aHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;

    var uri = Services.io.newURI("http://" + aHost, null, null);
    Services.perms.add(uri, "cookie", aPermission);
  },

  onContextMenuCopy: function(anItem) {
    var sClipboard = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper);
    var seln       = gCookiesTreeView.selection;
    var ci         = seln.currentIndex;
    var item       = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);

    if (item == null) {
      return;
    }

    if (anItem == "domain") {
      if (item.type == COOKIE_TYPE_CONTAINER) {
        sClipboard.copyString(item.rawHost);
      } else {
        sClipboard.copyString(item.host);
      }
    } else if (anItem == "name") {
      sClipboard.copyString(item.name);
    } else {
      sClipboard.copyString(item.value);
    }
  },

  onSort: function (aEvent, aSortInfos, anOrder) {
    var ctrlKey = false;
    if (aEvent != null && aEvent.button != 0) {
      // Not the right mouse button
      return;
    }
    if (aEvent != null) {
      if (typeof(aEvent.ctrlKey) == "boolean" || typeof(aEvent.metaKey) == "boolean") {
        ctrlKey = aEvent.ctrlKey || aEvent.metaKey;
      }
    }

    this.blockSelectEvent();
    gCookiesTreeView.saveState();

    if (typeof(anOrder) === "boolean") {
      var changeColumn = true;
      var ascending    = anOrder;
    } else {
      var changeColumn = (aSortInfos.column === this._lastSort.column) ? false : true;
      var ascending    = changeColumn ? true : !this._lastSortAscending;
    }

    if (aSortInfos.column === SORT_INFOS.HOST.column) {
      // Using control key while clicking on the column
      // allow to display the list in flat view.
      let viewMode = ctrlKey ? VIEW_MODE.FLAT : VIEW_MODE.TREE;
      changeColumn = gCookiesTreeView.viewMode != viewMode;
      gCookiesTreeView.viewMode = viewMode;
      if (changeColumn) {
        ascending = this._lastSortAscending;
      }
    } else {
      gCookiesTreeView.viewMode = VIEW_MODE.FLAT;
    }

    if (changeColumn) {
      gCookiesTreeView.sortInfos     = aSortInfos;
      gCookiesTreeView.sortAscending = ascending;
      gCookiesTreeView.refreshView(true);
    } else {
      gCookiesTreeView._treeBoxObj.beginUpdateBatch();
      gCookiesTreeView.selection.clearSelection();
      gCookiesTreeView.doSort(aSortInfos, ascending);
      gCookiesTreeView._treeBoxObj.invalidate();
      gCookiesTreeView.restoreState();
      gCookiesTreeView._treeBoxObj.endUpdateBatch();
    }

    // Adjust the columns sort indicators
    var oldCol          = document.getElementById(this._lastSort.column);
    var newCol          = document.getElementById(aSortInfos.column);
    var sortOrderString = ascending ? "ascending" : "descending";

    if (oldCol == null) {
      oldCol = document.getElementById(SORT_INFOS.DEFAULT.column);
      log.e("Old column unknown: ", this._lastSortColumn);
    }
    if (newCol == null) {
      newCol = document.getElementById(SORT_INFOS.DEFAULT.column);
      log.e("New column unknown: ", aSortInfos);
    }

    oldCol.removeAttribute("sortDirection");
    newCol.setAttribute("sortDirection", sortOrderString);

    this._lastSortAscending = ascending;
    this._lastSort          = aSortInfos;

    this.unblockSelectEvent();
    this.onItemSelected(null);
  },

  onSortByColNumber: function(aNumber) {
    var treeCols = document.getElementById("cookiesTreeCols").childNodes;
    var col      = null;
    var colNum   = 0;
    // Search for the first display column number:
    for (var i=0; i<treeCols.length; i++) {
      let obj = treeCols[i];
      if (obj.nodeName == "treecol" && obj.hidden == false) {
        if (colNum == aNumber) {
          col = obj;
          break;
        }
        colNum++;
      }
    }
    if (col != null) {
      this._tree.disabled = true;
      col.click();
      this._tree.disabled = false;
    }
  },

  onProtectionFilter: function(aFilterValue) {
    this.blockSelectEvent();
    gCookiesTreeView.saveState();

    var labelPFilter = document.getElementById("labelProtectionFilter");
    if (aFilterValue == "all") {
      gCookiesTreeView.globalFilter = FILTER.NONE;
      labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL);
    } else if (aFilterValue == "unprotected") {
      gCookiesTreeView.globalFilter = FILTER.UNPROTECTED;
      labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_UNPROT);
    } else {
      gCookiesTreeView.globalFilter = FILTER.PROTECTED;
      labelPFilter.value = this._bundle.getString(BUNDLE_PF_ALL_PROT);
    }

    this.unblockSelectEvent();

    // Update UI
    this.onItemSelected(null);
    this._tree.focus();
  },

  onFilter: function () {
    var searchValue = document.getElementById("textBoxSearch").value;

    this.blockSelectEvent();
    gCookiesTreeView.saveState();

    if (searchValue.length > 0) {
      var labelPFilter = document.getElementById("labelProtectionFilter");
      switch (gCookiesTreeView.globalFilter) {
        case FILTER.NONE:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER);
          break;
        case FILTER.PROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER_PROT);
          break;
        case FILTER.UNPROTECTED:
          labelPFilter.value = this._bundle.getString(BUNDLE_PF_FILTER_UNPROT);
          break;
      }
    }

    gCookiesTreeView.searchValue = searchValue;
    this.unblockSelectEvent();

    // Update UI
    this.onItemSelected(null);
  },

  focusFilterBox: function () {
    var filter = document.getElementById("textBoxSearch");
    filter.focus();
    filter.select();
  }

}; // var gCookieKeeper
