/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Public API: Private part, used by the bootstrap code too.
 *
 * ************************************************************************
 */

"use strict";

var CookieKeeperPrivate = {
  _pListLastSave : 0,
  _protectedList : null,

  init: function() {
    this._protectedList = {
      lastUpdate : 0,
      cookies    : [],
      storage    : []
    };
    this._pListLastSave = 0;
  },

  uninit: function() {
    this._pListLastSave = 0;
    this._protectedList = null;
  },

  getCookieWindows: function() {
    var e    = Services.wm.getEnumerator(null);
    var wins = [];

    while (e.hasMoreElements()) {
      let win = e.getNext().QueryInterface(Ci.nsIDOMChromeWindow);
      let id  = win.document.documentElement.getAttribute("id");

      for (let i=0; i<CK_WINDOWS_ID.length; i++) {
        if (CK_WINDOWS_ID[i] == id) {
          wins.push({id: id, window:win});
        }
      }
    }

    return wins;
  },

  _pStorageEquals: function(aItem) {
    // aItem can be a protected storage, or a dom storage
    return  aItem      !=  null        &&
            this.scope === aItem.scope &&
            this.key   === aItem.key;
  },

  makeProtectedStorage: function(aScope, aKey) {
    function getHost() {
      let host = this.scope.split(":")[0];
      return gCommon.reverseStr(host);
    }
    function getRawHost() {
      return gCommon.makeStrippedHost(this.getHost());
    }
    function getName() {
      return this.key;
    }

    var ps = {
      scope      : aScope,
      key        : aKey,
      equals     : CookieKeeperPrivate._pStorageEquals,
      getRawHost : getRawHost,
      getHost    : getHost,
      getName    : getName
    };

    return ps;
  },

  _pCookieEquals: function(aItem) {
    return  aItem     !=  null       &&
            this.host === aItem.host &&
            this.name === aItem.name &&
            this.path === aItem.path;
  },

  makeProtectedCookie: function(aHost, aName, aPath) {
    function getHost() {
      return this.host;
    }
    function getRawHost() {
      return gCommon.makeStrippedHost(this.host);
    }
    function getName() {
      return this.name;
    }
    let pc = {
      host       : aHost,
      name       : aName,
      path       : aPath,
      equals     : CookieKeeperPrivate._pCookieEquals,
      getRawHost : getRawHost,
      getHost    : getHost,
      getName    : getName
    };

    return pc;
  },

  /// Load the protection list, from the file: cookiekeeper.json (UTF-8 / JSON)
  /// In the user profile directory.
  /// @param doSync if true, dot not used async load.
  loadProtectedList: function(doSync) {

    function jsonFileLoaded(jsonString) {
      if (typeof(jsonString) != "string") {
        log.e("loadProtectedList.jsonFileLoaded: parameter is not a string", jsonString);
        return;
      }
      if (jsonString.length > 0) {
        let pLists = JSON.parse(jsonString);
        // Add the equals() function:
        let sList = pLists.cookies; // source
        let dList = CookieKeeperPrivate._protectedList.cookies; //destination
        for (let i=0; i<sList.length; i++) {
          let item = sList[i];
          if (typeof(item.host) == "string" && typeof(item.name) == "string" && typeof(item.path) == "string") {
            dList.push(CookieKeeperPrivate.makeProtectedCookie(item.host, item.name, item.path));
          } else {
            log.e("Found an invalid protected cookie at pos:" + i, item);
          }
        }
        sList = pLists.storage;
        dList = CookieKeeperPrivate._protectedList.storage;
        for (let i=0; i<sList.length; i++) {
          let item = sList[i];
          if (typeof(item.scope) == "string" && typeof(item.key) == "string") {
            dList.push(CookieKeeperPrivate.makeProtectedStorage(item.scope, item.key));
          } else {
            log.e("Found an invalid protected dom storage at pos:" + i, item);
          }
        }
        CookieKeeperPrivate._protectedList.lastUpdate = pLists.lastUpdate;

      } else {
        log.w("JSON file is empty:", jsonString);
        CookieKeeperPrivate._protectedList.lastUpdate = Date.now();
      }
      CookieKeeperPrivate._pListLastSave = Date.now();
    }

    function asyncLoaded(istream, status) {
      try {
        if (Components.isSuccessCode(status)) {
          let jsonString = NetUtil.readInputStreamToString(istream, istream.available(), {charset: "UTF-8"});
          jsonFileLoaded(jsonString);
        } else {
          log.e('error on file read isSuccessCode = ' + status);
        }
      } catch (ex) {
        log.e("loadProtectedList.asyncLoaded: " + ex, ex);
      }
    }

    function syncLoad(jsonFile) {
      try {
        let istream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
        istream.init(jsonFile, -1, -1, false);
        let jsonString = NetUtil.readInputStreamToString(istream, istream.available(), {charset: "UTF-8"});
        istream.close();

        jsonFileLoaded(jsonString);
      } catch (ex) {
        log.e("loadProtectedList.syncLoad: " + ex, ex);
      }
    }

    if (this._protectedList.lastUpdate != 0) {
      // Already loaded
      return;
    }

    let jsonFile = FileUtils.getFile("ProfD", [PROT_JSON_FILE], false);
    if (!jsonFile.exists()) {
      // Nothing to load
      this._protectedList.lastUpdate = Date.now();
      return;
    }

    if (doSync) {
      syncLoad(jsonFile);
    } else {
      NetUtil.asyncFetch(jsonFile, asyncLoaded);
    }
  },

  /// Save the protected cookies/storage to the file: cookiekeeper.json (UTF-8 / JSON)
  /// In the user profile directory
  /// @param doSync [boolean] if true, save in sync mode, if false, async.
  /// @note The sync mode should only be used during add-on shutdown
  saveProtectedList: function(doSync) {
    function writeDoneCallback(status) {
      if (!Components.isSuccessCode(status)) {
        // Handle error!
        log.e("cookiekeeper.json: error on write file (" + status + ")");
        return;
      }
      // Data has been written to the file.
      log.i("cookiekeeper.json: file written (ret status=" + status + ")");
      CookieKeeperPrivate._pListLastSave = Date.now();
    }

    if (this._pListLastSave < this._protectedList.lastUpdate) {
      // log.d("saveProtectedList: doSync=" + doSync);
      var openFlags  = FileUtils.MODE_WRONLY | FileUtils.MODE_TRUNCATE; // FileUtils.MODE_CREATE |
      var jsonFile   = FileUtils.getFile("ProfD", [PROT_JSON_FILE], false);
      var jsonString = JSON.stringify(this._protectedList);

      if (!jsonFile.exists()) {
        // To keep some privacy... not making the file readable by everybody
        jsonFile.create(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("600", 8)); // Unix file perm: -rw-------
      }
      var ostream = FileUtils.openFileOutputStream(jsonFile, openFlags);

      if (doSync) {
        let converter = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Components.interfaces.nsIConverterOutputStream);
        converter.init(ostream, "UTF-8", 0, 0);
        converter.writeString(jsonString);
        converter.close();
        this._pListLastSave = Date.now();
      } else {
        let converter = Cc['@mozilla.org/intl/scriptableunicodeconverter'].createInstance(Ci.nsIScriptableUnicodeConverter);
        converter.charset = 'UTF-8';
        let istream = converter.convertToInputStream(jsonString);
        NetUtil.asyncCopy(istream, ostream, writeDoneCallback);
      }
    } else { log.d("saveProtectedList: not needed"); }
  },

  /// @brief Clear only unwanted cookies
  clearUnwantedCookies: function() {
    var protList  = this._protectedList.cookies;
    var clearList = [];
    var cEnum     = sCookies.enumerator;

    function isProtectedCookie(aCookie) {
      var isProtected = false;
      for (let i=0; i<protList.length && !isProtected; i++) {
        let pi = protList[i];
        isProtected = pi.equals(aCookie);
      }

      return isProtected;
    }

    // Ususally it is not good to delete item on the same list we use an enumerator.
    // Since I don't know the implementation, I prefer to use a safe method:
    // Get all items we need to delete, then delete them.
    // This way, I'm sure to not messed up the iterator.
    while (cEnum.hasMoreElements()) {
      let cookie = cEnum.getNext();
      if (cookie && cookie instanceof Ci.nsICookie2 && !isProtectedCookie(cookie)) {
        clearList.push(cookie);
      }
    }

    for (let i=0; i<clearList.length; i++) {
      let cookie = clearList[i];
      sCookies.remove(cookie.host, cookie.name, cookie.path, false);
    }
  },

  clearUnwantedDomStorage: function() {
    const scopeAndKey = "(scope=? AND key=?)";
    var delQuery      = "DELETE FROM " + DOM_DB_TABLE + " WHERE NOT (";
    var protList      = this._protectedList.storage;
    var protInfo      = [];
    var connection    = null;

    // Get protected DOM storage infos in a single list
    for (let i=0; i<protList.length; i++) {
      let pi = protList[i];
      protInfo.push(pi.scope);
      protInfo.push(pi.key);
    }
    // Build the query

    try {
      // Perform the clean sync, not using Sqlite.jsm, because
      // it will not be executed on shutdown.
      // This module will be unregistered, before the AsyncShutdown task is launch.

      let path   = OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME);
      let file   = FileUtils.File(path);
      connection = Services.storage.openDatabase(file);

      if (protInfo.length > 0) {
        for (let i=0; i<protInfo.length; i+=2) {
          if (i > 0) {
            delQuery = delQuery + " or " + scopeAndKey;
          } else {
            delQuery = delQuery + scopeAndKey;
          }
        }
        delQuery = delQuery + ")";

        if (connection.connectionReady) {
          let stmt   = connection.createStatement(delQuery);

          for (let i=0; i<protInfo.length; i++) {
            stmt.bindByIndex(i, protInfo[i]);
          }
          stmt.execute();
          stmt.finalize();
        }
      } else {
        log.d("No protected DOM storage found, truncate the table");
        let path   = OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME);
        let file   = FileUtils.File(path);
        connection = Services.storage.openDatabase(file);
        if (connection.connectionReady) {
          connection.executeSimpleSQL("DELETE FROM " + DOM_DB_TABLE);
        }
      }
    } catch (ex) {
      log.e("Error cleaning dom storage:", ex);
    } finally {
      if (connection != null) {
        // log.d("Closing " + DOM_DB_NAME + " connection");
        connection.close();
        sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_CLEAN);
      }
    }

  } // clearUnwantedDomStorage

}; // CookieKeeperPrivate
