/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Preferences: Show/remove protected cookies list.
 *
 * ************************************************************************
 */

"use strict";

var gCookieKeeperProtCookiesPrefs = {
  _strCookie    : "Cookie",
  _strDomStorage: "DOM storage",
  _buttonRemove : null,
  _treeView     : null,
  _treeChildren : null,
  _treeCols     : [],
  _pItems       : [],
  _sortColumn   : 0,
  _sortOrderAsc : false,

  init: function() {
    var bundle         = document.getElementById("bundlePreferences");
    this._buttonRemove = document.getElementById("buttonRemove");
    this._treeView     = document.getElementById("treeProtectedItem");
    this._treeChildren = document.getElementById("treeProtItemsChildren");

    this._treeCols.push(document.getElementById("tiType"));
    this._treeCols.push(document.getElementById("tiDomain"));
    this._treeCols.push(document.getElementById("tiName"));
    this._treeCols.push(document.getElementById("tiPath"));

    this._strCookie     = bundle.getString("type_cookie");
    this._strDomStorage = bundle.getString("type_dom_storage");

    let pItems = CookieKeeper.getProtectedItems();
    this._pItems = pItems.cookies.concat(pItems.storage);

    if (this._pItems.length == 0) {
      this._buttonRemove.disabled = true;
      this._treeView.disabled     = true;
    } else {
      this._treeView.seltype = "single";
      this.onSort(null, 0);
      this._treeView.view.selection.select(0);
    }
  },

  _fillList: function() {

    for (var i=0; i<this._pItems.length; i++) {
      let pc   = this._pItems[i];
      let item = document.createElement("treeitem");
      let row  = document.createElement("treerow");

      if (typeof(pc.scope) == "string") {
        // DOM Storage
        let cell = document.createElement("treecell");
        cell.setAttribute("label", this._strDomStorage);
        row.appendChild(cell);

        let scope = pc.scope.split(":");
        let host  = gCommon.reverseStr(scope[0]);
        cell = document.createElement("treecell");
        cell.setAttribute("label", host + " (" + scope[1] + ":" + scope[2] + ")");
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.key);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  "");
        row.appendChild(cell);

      } else {
        // Cookie
        let cell = document.createElement("treecell");
        cell.setAttribute("label", this._strCookie);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label", pc.host);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.name);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.path);
        row.appendChild(cell);
      }

      item.appendChild(row);
      item.setAttribute("accessibleType", "101F")
      this._treeChildren.appendChild(item);

      // To easily remove this item when a protected cookie is deleted
      pc.view = item;

    }

  },

  onRemove: function() {
    var seln = this._treeView.view.selection;
    var selIndexes = [];

    if (seln.count == 1) {
      selIndexes.push(seln.currentIndex);
    } else if (seln.count > 1) {
      let rangeCount = seln.getRangeCount();
      for (let i = 0; i < rangeCount; i++) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);
        for (let j = min.value; j <= max.value; j++) {
          selIndexes.push(j);
        }
      }
    }

    for (let i=selIndexes.length-1; i>=0; i--) {
      let index = selIndexes[i];
      var pCookie = this._pItems[index];

      this._pItems.splice(index, 1);

      // Remove protected cookie:
      CookieKeeper.removeProtectedItem(pCookie);

      // Update ui:
      this._treeChildren.removeChild(pCookie.view);
      var rowCount = this._pItems.length;
      if (rowCount == 0) {
        this._buttonRemove.disabled = true;
        this._treeView.disabled     = true;
      } else {
        var newIndex = (index < rowCount ? index : rowCount - 1);
        seln.select(newIndex);
      }
    }

    this._treeView.focus();
  },

  // Sort the _pItems list
  _sort: function(a, b) {
    var c = 0;
    var loop = 0|0; // prevent infinte loop... we never know
    const T_COOKIE = 0;
    const T_STORE  = 1;
    let ta  = typeof(a.scope) == "string" ? T_STORE : T_COOKIE;
    let tb  = typeof(b.scope) == "string" ? T_STORE : T_COOKIE;

    let sortColumn = gCookieKeeperProtCookiesPrefs._sortColumn;

    while (c == 0 && loop < 2) {
      loop++;
      switch (sortColumn) {
        case 0: // Type (Cookie/Dom storage)
          c = ta-tb;
          if (c != 0) { break; }
          // fall through
        case 1: { // Host/Domain
          c = gCommon.cmpHostName(a.getHost(), b.getHost());
          if (c != 0) { break; }
        } // fall through
        case 2: { // Name/Key
          let na  = ta == T_COOKIE ? a.name : a.key;
          let nb  = tb == T_COOKIE ? b.name : b.key;
          c = a.getName().toLowerCase().localeCompare(b.getName().toLowerCase());
          if (c != 0) { break; }
        } // fall through
        case 3: { // Path
          let pa  = ta == T_COOKIE ? a.path : "";
          let pb  = tb == T_COOKIE ? b.path : "";
          c = pa.toLowerCase().localeCompare(pb.toLowerCase());
         } // fall through
        default:
          if (c == 0 && sortColumn>0) { sortColumn = 0; }
      }
    }

    if (gCookieKeeperProtCookiesPrefs._sortOrderAsc) {
      return c;
    } else {
      return -c;
    }
  },

  onSort: function(aEvent, aColumn) {
    if ((aEvent != null && aEvent.button != 0) || aColumn < 0 || aColumn >= this._treeCols.length) {
      // param check (not usefull, but to be sure)
      return;
    }

    var colHeader = this._treeCols[aColumn];
    var column    = null;

    if (this._sortColumn == aColumn) {
      this._sortOrderAsc = !this._sortOrderAsc;
    } else {
      var oldColHeader = this._treeCols[this._sortColumn];

      this._sortOrderAsc = true;
      this._sortColumn   = aColumn;

      oldColHeader.removeAttribute("sortDirection");
    }
    colHeader.setAttribute("sortDirection", this._sortOrderAsc ? "ascending" : "descending");

    try {
      this._pItems.sort(this._sort);
    } catch (ex) {
      log.e(ex);
    }

    // Not pretty, but easy to do
    while (this._treeChildren.firstChild) {
      this._treeChildren.removeChild(this._treeChildren.firstChild);
    }

    this._fillList();
  }
};
