/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Permissions panel.
 * Original code from browser/components/preferences/permissions.js (Mozilla Firefox 30.0a1 source code)
 *
 * ************************************************************************
 */

"use strict";

const NOTIFICATION_FLUSH_PERMISSIONS = "flush-pending-permissions";
const PERMISSION_TYPE                = "cookie";

function Permission(host, rawHost, type, capability, capabilityId) {
  this.host         = host;
  this.rawHost      = rawHost;
  this.type         = type;
  this.capability   = capability;
  this.capabilityId = capabilityId;
}

function permissionsComparator(a, b) {
  return a.toLowerCase().localeCompare(b.toLowerCase());
}

var gCookieKeeperPermissionsPrefs = {
  _permissions  : [],
  _bundle       : null,
  _tree         : null,
  _lastPermissionSortColumn: "siteCol",
  _lastPermissionSortAscending: false,

  _view: {
    _rowCount: 0,
    get rowCount() {
      return this._rowCount;
    },

    getCellText: function(aRow, aColumn) {
      if (aColumn.id == "siteCol") {
        return gCookieKeeperPermissionsPrefs._permissions[aRow].rawHost;
      } else if (aColumn.id == "statusCol") {
        return gCookieKeeperPermissionsPrefs._permissions[aRow].capability;
      }
      return "";
    },

    isSeparator: function(aIndex) { return false; },
    isSorted: function() { return false; },
    isContainer: function(aIndex) { return false; },
    setTree: function(aTree) {},
    getImageSrc: function(aRow, aColumn) {},
    getProgressMode: function(aRow, aColumn) {},
    getCellValue: function(aRow, aColumn) {},
    cycleHeader: function(aColumn) {},
    getRowProperties: function(aRow){ return ""; },
    getColumnProperties: function(aColumn) { return ""; },
    getCellProperties: function(aRow,aColumn) {
      if (aColumn.element.getAttribute("id") == "siteCol") {
        return "ltr";
      }
      return "";
    }
  },

  init: function() {
    this._bundle = document.getElementById("bundlePreferences");

    var os = Components.classes["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    os.notifyObservers(null, NOTIFICATION_FLUSH_PERMISSIONS, PERMISSION_TYPE);
    os.addObserver(this, "perm-changed", false);

    this._loadPermissions();

    var urlField = document.getElementById("url");
    urlField.focus();
  },

  uninit: function() {
    var os = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    os.removeObserver(this, "perm-changed");
  },

  _getCapabilityString: function(aCapability) {
    var stringKey = null;
    switch (aCapability) {
    case MOZ_COOKIE_PERM_ALLOW:
      stringKey = "can";
      break;
    case MOZ_COOKIE_PERM_DENY:
      stringKey = "cannot";
      break;
    case MOZ_COOKIE_PERM_FIRST:
      stringKey = "canAccessFirstParty";
      break;
    case MOZ_COOKIE_PERM_SESSION:
      stringKey = "canSession";
      break;
    }
    return this._bundle.getString(stringKey);
  },

  addPermission: function(aCapability) {

    var textbox = document.getElementById("url");
    var host = textbox.value.replace(/^\s*([-\w]*:\/+)?/, ""); // trim any leading space and scheme
    try {
      var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
      var uri = ioService.newURI("http://"+host, null, null);
      host = uri.host;
    } catch(ex) {
      var promptService = Cc["@mozilla.org/embedcomp/prompt-service;1"].getService(Ci.nsIPromptService);
      var message = this._bundle.getString("invalidURI");
      var title = this._bundle.getString("invalidURITitle");
      promptService.alert(window, title, message);
      return;
    }

    // check whether the permission already exists, if not, add it
    var exists = false;
    for (let i=0; i<this._permissions.length; i++) {
      if (this._permissions[i].rawHost == host) {
        // Avoid calling the permission manager if the capability settings are
        // the same. Otherwise allow the call to the permissions manager to
        // update the listbox for us.
        exists = this._permissions[i].capabilityId == aCapability;
        break;
      }
    }

    if (!exists) {
      host = (host.charAt(0) == ".") ? host.substring(1,host.length) : host;
      let uri = ioService.newURI("http://" + host, null, null);
      Services.perms.add(uri, PERMISSION_TYPE, aCapability);
    }
    textbox.value = "";
    textbox.focus();
    this.onHostInput(textbox);

    // enable "remove all" button as needed
    document.getElementById("removeAllPermissions").disabled = this._permissions.length == 0;
  },

  onHostInput: function(aSiteField) {
    document.getElementById("btnSession").disabled = !aSiteField.value;
    document.getElementById("btnBlock").disabled = !aSiteField.value;
    document.getElementById("btnAllow").disabled = !aSiteField.value;
  },

  onHostKeyPress: function(aEvent) {
    if (aEvent.keyCode == KeyEvent.DOM_VK_RETURN)
      document.getElementById("btnAllow").click();
  },

  observe: function(aSubject, aTopic, aData) {
    if (aTopic == "perm-changed") {
      var permission = aSubject.QueryInterface(Ci.nsIPermission);
      if (aData == "added") {
        this._addPermissionToList(permission);
        this._view._rowCount++;
        this._tree.treeBoxObject.rowCountChanged(this._view.rowCount - 1, 1);
        // Re-do the sort, since we inserted this new item at the end.
        gTreeUtils.sort(this._tree, this._view, this._permissions,
                        this._lastPermissionSortColumn,
                        permissionsComparator,
                        this._lastPermissionSortColumn,
                        !this._lastPermissionSortAscending);
      } else if (aData == "changed") {
        for (let i=0; i<this._permissions.length; i++) {
          if (this._permissions[i].host == permission.host) {
            this._permissions[i].capability   = this._getCapabilityString(permission.capability);
            this._permissions[i].capabilityId = permission.capability;
            break;
          }
        }
        // Re-do the sort, if the status changed from Block to Allow
        // or vice versa, since if we're sorted on status, we may no
        // longer be in order.
        if (this._lastPermissionSortColumn.id == "statusCol") {
          gTreeUtils.sort(this._tree, this._view, this._permissions,
                          this._lastPermissionSortColumn,
                          permissionsComparator,
                          this._lastPermissionSortColumn,
                          !this._lastPermissionSortAscending);
        }
        this._tree.treeBoxObject.invalidate();
      }
      // No UI other than this window causes this method to be sent a "deleted"
      // notification, so we don't need to implement it since Delete is handled
      // directly by the Permission Removal handlers. If that ever changes, those
      // implementations will have to move into here.
    }
  },

  onPermissionSelected: function() {
    var hasSelection = this._tree.view.selection.count > 0;
    var hasRows = this._tree.view.rowCount > 0;
    document.getElementById("removePermission").disabled = !hasRows || !hasSelection;
    document.getElementById("removeAllPermissions").disabled = !hasRows;
  },

  onPermissionDeleted: function() {
    if (!this._view.rowCount) {
      return;
    }
    var removedPermissions = [];
    gTreeUtils.deleteSelectedItems(this._tree, this._view, this._permissions, removedPermissions);
    for (let i=0; i<removedPermissions.length; i++) {
      var p = removedPermissions[i];
      Services.perms.remove(p.host, p.type);
    }
    document.getElementById("removePermission").disabled = !this._permissions.length;
    document.getElementById("removeAllPermissions").disabled = !this._permissions.length;
  },

  onAllPermissionsDeleted: function() {
    if (!this._view.rowCount) {
      return;
    }
    var removedPermissions = [];
    gTreeUtils.deleteAll(this._tree, this._view, this._permissions, removedPermissions);
    for (let i=0; i<removedPermissions.length; i++) {
      var p = removedPermissions[i];
      Services.perms.remove(p.host, p.type);
    }
    document.getElementById("removePermission").disabled = true;
    document.getElementById("removeAllPermissions").disabled = true;
  },

  onPermissionKeyPress: function(aEvent) {
    if (aEvent.keyCode == aEvent.DOM_VK_DELETE) {
      this.onPermissionDeleted();
    }
  },

  onPermissionSort: function(aEvent, aColumn) {
    if (aEvent != null && aEvent.button != 0) {
      return;
    }
    function getColumnElem(_col) {
      let ce = null;
      if (typeof(_col) == "undefined" || _col == null) {
      } else if (_col == "rawHost") {
        ce = document.getElementById("siteCol");
      } else if (_col == "capability") {
        ce = document.getElementById("statusCol");
      }
      return ce;
    }

    var columnElem = getColumnElem(this._lastPermissionSortColumn);
    if (columnElem != null) {
      columnElem.removeAttribute("sortDirection");
    }
    this._lastPermissionSortAscending = gTreeUtils.sort(this._tree,
                                                        this._view,
                                                        this._permissions,
                                                        aColumn,
                                                        permissionsComparator,
                                                        this._lastPermissionSortColumn,
                                                        this._lastPermissionSortAscending);
    this._lastPermissionSortColumn = aColumn;
    columnElem = getColumnElem(aColumn);
    if (columnElem != null) {
      columnElem.setAttribute("sortDirection", this._lastPermissionSortAscending ? "ascending" : "descending");
    }
  },

  _loadPermissions: function() {
    this._tree = document.getElementById("treePermissions");
    this._permissions = [];

    // load permissions into a table
    var enumerator = Services.perms.enumerator;
    while (enumerator.hasMoreElements()) {
      let nextPermission = enumerator.getNext().QueryInterface(Ci.nsIPermission);
      this._addPermissionToList(nextPermission);
    }

    this._view._rowCount = this._permissions.length;

    // sort and display the table
    this._tree.treeBoxObject.view = this._view;
    this.onPermissionSort(null, "rawHost");

    // disable "remove all" button if there are none
    document.getElementById("removeAllPermissions").disabled = this._permissions.length == 0;
  },

  _addPermissionToList: function(aPermission) {
    if (aPermission.type == PERMISSION_TYPE) {

      var host = aPermission.host;
      var capabilityString = this._getCapabilityString(aPermission.capability);
      var p = new Permission(host,
                             (host.charAt(0) == ".") ? host.substring(1,host.length) : host,
                             aPermission.type,
                             capabilityString,
                             aPermission.capability);
      this._permissions.push(p);
    }
  },

  // ***********************
  // Context menu functions:

  onContextMenu: function(aEvent) {
    var seln            = this._view.selection;
    var showContextMenu = this._permissions.length > 0 && this._view.rowCount > 0 && seln.count > 0;
    if (!showContextMenu) {
      aEvent.preventDefault();
    }
    return aEvent;
  },

  _lastSelectedItems: [],
  onShowContextCookieMenu: function(aEvent) {
    var seln            = this._view.selection;
    var mpccPermAllow   = document.getElementById("mpcc-perm-allow");
    var mpccPermSession = document.getElementById("mpcc-perm-session");
    var mpccPermDeny    = document.getElementById("mpcc-perm-deny");

    this._lastSelectedItems = [];

    if (seln.currentIndex == -1 || seln.count == 0) {
      return false;
    }

    mpccPermAllow.hidden   = false;
    mpccPermSession.hidden = false;
    mpccPermDeny.hidden    = false;

    // Hide unusefull menu items
    if (seln.count > 1) {
      let nPermAllow   = 0;
      let nPermSession = 0;
      let nPermDeny    = 0;
      let nTotal       = 0;
      let rangeCount   = seln.getRangeCount();

      for (let i = 0; i < rangeCount; i++) {
        let min = {}; var max = {};
        seln.getRangeAt(i, min, max);

        for (let j = min.value; j <= max.value; j++) {
          nTotal++;
          let item = gCookieKeeperPermissionsPrefs._permissions[j];
          this._lastSelectedItems.push(item);

          switch (item.capabilityId) {
          case MOZ_COOKIE_PERM_ALLOW:
            nPermAllow++;
            break;
          case MOZ_COOKIE_PERM_SESSION:
            nPermSession++;
            break;
          case MOZ_COOKIE_PERM_DENY:
            nPermDeny++;
            break;
          }
        }
      }

      if (nTotal == nPermAllow) {
        mpccPermAllow.hidden = true;
      } else if (nTotal == nPermSession) {
        mpccPermSession.hidden = true;
      } else if (nTotal == nPermDeny) {
        mpccPermDeny.hidden = true;
      }

    } else {
      let item = gCookieKeeperPermissionsPrefs._permissions[seln.currentIndex];
      this._lastSelectedItems.push(item);

      switch (item.capabilityId) {
      case MOZ_COOKIE_PERM_ALLOW:
        mpccPermAllow.hidden = true;
        break;
      case MOZ_COOKIE_PERM_SESSION:
        mpccPermSession.hidden = true;
        break;
      case MOZ_COOKIE_PERM_DENY:
        mpccPermDeny.hidden = true;
        break;
      }
    }

    return true;
  },

  onContextMenuPermission: function(aPerm) {
    var ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);

    for (let i=0; i<this._lastSelectedItems.length; i++) {
      let item = this._lastSelectedItems[i];
      let uri = ioService.newURI("http://" + item.host, null, null);
      Services.perms.add(uri, PERMISSION_TYPE, aPerm);
    }
  },

  onContextMenuCopy: function() {
    var sClipboard = Cc["@mozilla.org/widget/clipboardhelper;1"].getService(Ci.nsIClipboardHelper);
    var seln       = this._view.selection;
    var ci         = seln.currentIndex;
    var rawHost    = ci == -1 ? null : gCookieKeeperPermissionsPrefs._permissions[ci].rawHost;

    if (rawHost == null) {
      return;
    }
    sClipboard.copyString(rawHost);
  }
};
