/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Public API
 *
 * ************************************************************************
 */

"use strict";

var EXPORTED_SYMBOLS = ["CookieKeeper"];

Components.utils.import("resource://gre/modules/osfile.jsm");
Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/devtools/Console.jsm");
Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://gre/modules/Task.jsm");
Components.utils.import("resource://gre/modules/FileUtils.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");

let sPrefs    = Services.prefs;
let sObserver = Services.obs;
let sCookies  = Services.cookies;
let self      = this;

Services.scriptloader.loadSubScript("chrome://cookiekeeper/content/common.js",     self);
Services.scriptloader.loadSubScript("chrome://cookiekeeper/content/ck.private.js", self);

/**
 * @brief Common/Public CookieKeeper function.
 *
 * @par Observer example:
 * see https://developer.mozilla.org/en-US/Add-ons/Overlay_Extensions/XUL_School/Observer_Notifications
 ~~~~~~~~~~~~~{.js}
    let pCObserver = {
      observe : function(aSubject, aTopic, aData) {
          if (aTopic == CookieKeeper.OBS_TOPIC) {
            if (aData == CookieKeeper.OBS_STARTUP || aData == CookieKeeper.OBS_SHUTDOWN) {

            } else {
              var pCookie = aSubject.wrappedJSObject.pcookie;
              var action  = "added";
              if (aData == CookieKeeper.OBS_DEL) {
                action = "removed";
              }
              window.alert("PCookie " + action + ": " + pCookie.host + ":" + pCookie.name);
          }
        }
      }
 ~~~~~~~~~~~~~
 */

let timerSave = null;

function saveProtectedListTimer() {
  var timerCallback = {
    notify: function(timer) {
      timerSave = null;
      CookieKeeperPrivate.saveProtectedList(false);
    }
  }
  if (timerSave == null) {
    timerSave = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    timerSave.initWithCallback(timerCallback, 2000, Ci.nsITimer.TYPE_ONE_SHOT);
  } else {
    timerSave.cancel();
    timerSave.initWithCallback(timerCallback, 2000, Ci.nsITimer.TYPE_ONE_SHOT);
  }
}

var CookieKeeper = {
  get API_VERSION()  { return 2; },                       /// The CookieKeeper public API version
  get OBS_TOPIC()    { return CK_OBS_PCOOKIE_TOPIC; },    /// Observer ProtectedCookies topic
  get OBS_STARTUP()  { return CK_OBS_PCOOKIE_STARTUP; },  /// Add-on is starting
  get OBS_SHUTDOWN() { return CK_OBS_PCOOKIE_SHUTDOWN; }, /// Add-on is shutting down
  get OBS_ADD()      { return CK_OBS_PCOOKIE_ADD; },      /// A protected cookie is added
  get OBS_DEL()      { return CK_OBS_PCOOKIE_DEL; },      /// A protected cookie is removed

  // Preferences panel open options:
  get PREFPANE_OPTIONS()  { return CK_PREFPANE_OPTIONS; },      /// To show the CookieKeeper preference pane
  get PREFPANE_PCOOKIES() { return CK_PREFPANE_PCOOKIES; },     /// To show the protected cookies pane
  get PREFPANE_BROWSER()  { return CK_PREFPANE_BROWSER; },      /// To show the browser preferences pane
  get PREFPANE_PERMS()    { return CK_PREFPANE_PERMISSIONS; },  /// To show the cookies permissions pane

  get COOKIE_TYPE_CONTAINER() { return COOKIE_TYPE_CONTAINER; }, // A container (directory/folder)
  get COOKIE_TYPE_STANDARD()  { return COOKIE_TYPE_STANDARD;  }, // A standard cookie
  get COOKIE_TYPE_STORAGE()   { return COOKIE_TYPE_STORAGE;   }, // DOM Storage

  /**
   * @brief Called only by the CookieKeeper bootstrap startup() method
   */
  init: function() {
    CookieKeeperPrivate.init();
    CookieKeeperPrivate.loadProtectedList(false);
    sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_STARTUP);
  },

  /**
   * @brief Called only by the CookieKeeper bootstrap shutdown() method
   */
  uninit: function() {
    if (timerSave != null) {
      timerSave.cancel();
      timerSave = null;
    }
    CookieKeeperPrivate.saveProtectedList(true);
    CookieKeeperPrivate.uninit();
    sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_SHUTDOWN);

    sPrefs    = null;
    sObserver = null;
    sCookies  = null;
    self      = null;
    CookieKeeperPrivate = null;
  },

  /**
   * @brief Show the main CookieKeeper window. If the dialogue is already opened, it will be focus.
   * @param window The DOM window.
   */
  showMainWindow: function(window) {
    var ckws = CookieKeeperPrivate.getCookieWindows();
    var opened = false;
    for (let i=0; i<ckws.length && opened == false; i++) {
      let ckw = ckws[i];
      if (ckw.id == CK_WINDOWS_ID[CK_WINDOW_MAIN_IDX]) {
        ckw.window.focus();
        opened = true;
      }
    }

    if (!opened) {
      window.open(
          "chrome://cookiekeeper/content/cookiekeeper.xul",
          "cookieKeeperWindow",
          "chrome,toolbar,centerscreen,resizable"
      );
    }
  },

  /**
   * @brief Show the CookieKeeper preferences dialogue. If the window is already opened, it will be focus.
   * @param window A DOM window.
   * @param aPane The pane to show (PREFPANE_OPTIONS/PREFPANE_PCOOKIES/PREFPANE_PERMS/PREFPANE_BROWSER)
   * if null, show the last pane.
   * @note If the window is already opened, the aPane argument will not be honored.
   */
  showPreferencesDialogue: function(window, aPane) {
    var ckws = CookieKeeperPrivate.getCookieWindows();
    var opened = false;
    for (let i=0; i<ckws.length && opened == false; i++) {
      let ckw = ckws[i];
      if (ckw.id == CK_WINDOWS_ID[CK_WINDOW_PREFS_IDX]) {
        ckw.window.focus();
        opened = true;
      }
    }

    if (!opened) {
        let xulRuntime = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULRuntime);
        let osOptions = "";
        if (xulRuntime.OS == "WINNT") {
          // MS Windows love modal dialogue
          osOptions = ",modal";
        } else if (xulRuntime.OS == "Darwin") {
          // Don't really know why, but recommanded by Mozilla on MacOSX
          osOptions = ",dialog=yes";
        } // else {} // GNU/Linux don't like restrictions ;-)
        window.openDialog("chrome://cookiekeeper/content/prefs.main.xul",
                          "cookieKeeperPrefsWindow",
                          "chrome,titlebar,toolbar,centerscreen" + osOptions,
                          aPane);
    }
  },

  /**
   * @brief Close all CookieKeeper windows.
   */
  closeCookieWindow: function() {
    function timeoutClose(aWin) {
      try {
        aWin.close();
        if (!aWin.closed) {
          aWin.setTimeout(function() { timeoutClose(aWin); }, 50);
        }
      } catch (ex) {
        Components.utils.reportError(ex);
      }
    }

    var ckws = CookieKeeperPrivate.getCookieWindows();
    for (let i=ckws.length-1; i>=0; i--) {
      let ckw = ckws[i];
      timeoutClose(ckw.window);
    }
  },

  /**
   * @brief Get the protected items
   * @return an object with two properties (array):
   * protected cookies (cookies) and protected storage (storage)
   * { cookies : [], storage : [] }
   */
  getProtectedItems: function() {
    let protectedList = { cookies : [], storage : [] }
    let protList      = CookieKeeperPrivate._protectedList.cookies;
    let pll           = protList.length;
    for (let i=0; i<pll; i++) {
      let pi = protList[i];
      let pc = CookieKeeperPrivate.makeProtectedCookie(pi.host, pi.name, pi.path);
      protectedList.cookies.push(pc);
    }

    protList = CookieKeeperPrivate._protectedList.storage;
    pll      = protList.length;
    for (let i=0; i<pll; i++) {
      let pi = protList[i];
      let ps = CookieKeeperPrivate.makeProtectedStorage(pi.scope, pi.key);
      protectedList.storage.push(ps);
    }

    return protectedList;
  },

  /**
   * @brief Test if this item is protected
   * @param aItem an item (cookie or dom storage)
   * @return true if the item is protected
   * @note
   * - If aItem is a cookie, it may have the property "type" set to COOKIE_TYPE_STANDARD.
   * - If aItem is a cookie, it must have host/name/path properties (string)
   * - If aItem is a dom storage, it must have the property "type" set to COOKIE_TYPE_STORAGE.
   * - If aItem is a dom storage, it must have scope/name properties (string)
   */
  isProtectedItem: function(aItem) {
    var protList    = [];
    var isProtected = false;

    switch (aItem.type) {
    case COOKIE_TYPE_STANDARD:
      protList = CookieKeeperPrivate._protectedList.cookies;
      break;
    case COOKIE_TYPE_STORAGE:
      protList = CookieKeeperPrivate._protectedList.storage;
      break;
    default:
      log.e("Unknown cookie type " + aItem.type);
      break;
    }

    for (let i=0; i<protList.length && !isProtected; i++) {
      let pi = protList[i];
      isProtected = pi.equals(aItem);
    }

    return isProtected;
  },

  /**
   * @brief Add a protected item to the list
   * @param aItem A protected item (cookie: host/name/path; storage: scope/key)
   * @return true if the protected item has been added
   * @par
   * Notify the observer: OBS_TOPIC/OBS_ADD
   */
  addProtectedItem: function(aItem) {
    if (aItem == null) {
      log.e("addProtectedItem: Item is not a valid cookie item");
      return false;
    }

    var protList    = [];
    var isProtected = false;
    var type        = -1;
    if (typeof(aItem.type) != "number") {
      if (typeof(aItem.host) == "string" && typeof(aItem.name) == "string" && typeof(aItem.path) == "string") {
        type = COOKIE_TYPE_STANDARD;
      } else if (typeof(aItem.scope) == "string" && typeof(aItem.key) == "string") {
        type = COOKIE_TYPE_STORAGE;
      }
    } else {
      type = aItem.type;
    }

    switch (type) {
    case COOKIE_TYPE_STANDARD:
      protList = CookieKeeperPrivate._protectedList.cookies;
      break;
    case COOKIE_TYPE_STORAGE:
      protList = CookieKeeperPrivate._protectedList.storage;
      break;
    default:
      log.e("addProtectedItem: item is invalid", aItem);
      return false;
    }

    for (let i=0; i<protList.length && !isProtected; i++) {
      let pi = protList[i];
      isProtected = pi.equals(aItem);
    }

    if (!isProtected) {
      // Create two protected item, to be sure to not return an iten in our list
      // If the item is modified by an observer, it won't modify ours.
      let pi   = {};
      let data = {};
      switch (type) {
      case COOKIE_TYPE_STANDARD:
        let pc = CookieKeeperPrivate.makeProtectedCookie(aItem.host, aItem.name, aItem.path);
        pi     = CookieKeeperPrivate.makeProtectedCookie(aItem.host, aItem.name, aItem.path);
        data   = {pcookie : pc, pstorage : null};
        break;
      case COOKIE_TYPE_STORAGE:
        let ps = CookieKeeperPrivate.makeProtectedStorage(aItem.scope, aItem.key);
        pi     = CookieKeeperPrivate.makeProtectedStorage(aItem.scope, aItem.key);
        data   = {pcookie : null, pstorage : ps};
        break;
      }

      protList.push(pi);
      CookieKeeperPrivate._protectedList.lastUpdate = Date.now();
      saveProtectedListTimer();

      let subject = { wrappedJSObject: data };
      sObserver.notifyObservers(subject, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_ADD);
    } else {
      log.w("Protected item already added: ", aItem);
    }

    return !isProtected;
  },

  /**
   * @brief Remove a protected item
   * @param aItem A protected item (cookie: host/name/path; storage: scope/key)
   * @return true if the protected item has been removed
   * @par
   * Notify the observer: OBS_TOPIC/OBS_DEL
   */
  removeProtectedItem: function(aItem) {
    if (aItem == null) {
      log.e("removeProtectedItem: Item is null");
      return false;
    }

    var protList = [];
    var exists   = false;
    var type     = -1;
    if (typeof(aItem.type) != "number") {
      if (typeof(aItem.host) == "string" && typeof(aItem.name) == "string" && typeof(aItem.path) == "string") {
        type = COOKIE_TYPE_STANDARD;
      } else if (typeof(aItem.scope) == "string" && typeof(aItem.key) == "string") {
        type = COOKIE_TYPE_STORAGE;
      }
    } else {
      type = aItem.type;
    }

    switch (type) {
    case COOKIE_TYPE_STANDARD:
      protList = CookieKeeperPrivate._protectedList.cookies;
      break;
    case COOKIE_TYPE_STORAGE:
      protList = CookieKeeperPrivate._protectedList.storage;
      break;
    default:
      log.e("addProtectedItem: item is invalid", aItem);
      return false;
    }

    for (let i=0; i<protList.length; i++) {
      let pi = protList[i];
      if (pi.equals(aItem)) {
        exists = true;
        protList.splice(i, 1);
        CookieKeeperPrivate._protectedList.lastUpdate = Date.now();
        saveProtectedListTimer();

        let data = { pcookie : null, pstorage : null };
        switch (type) {
        case COOKIE_TYPE_STANDARD:
          data.pcookie = pi;
          break;
        case COOKIE_TYPE_STORAGE:
          data.pstorage = pi;
          break;
        }

        let subject = { wrappedJSObject: data };
        sObserver.notifyObservers(subject, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DEL);

        i = protList.length;
      }
    }

    if (DEBUG && !exists) {
      log.w("Protected item do not exists: ", aItem);
    }

    return exists;
  },

  /**
   * Clear unprotected cookies/dom storage.
   * @param atExit Boolean, must be true only when the mozilla application is closing.
   *  - If true, read the preferences to know if we need to clear cookies/dom storage.
   *  - If false, clear unprotected cookies and dom storage.
   */
  clearAllCookies: function(atExit) {
    var clearCookies = true;
    var clearStorage = true;

    if (atExit == true) {
      clearCookies = sPrefs.prefHasUserValue(PREF_CLEAR_COOKIES_AT_EXIT) ? sPrefs.getBoolPref(PREF_CLEAR_COOKIES_AT_EXIT) : false;
      clearStorage = sPrefs.prefHasUserValue(PREF_CLEAR_DOM_AT_EXIT)     ? sPrefs.getBoolPref(PREF_CLEAR_DOM_AT_EXIT)     : false;
    }

    if (clearCookies) {
      log.d("Clear unprotected cookies.");
      CookieKeeperPrivate.clearUnwantedCookies();
    }
    if (clearStorage) {
      log.d("Clear unprotected storage.");
      CookieKeeperPrivate.clearUnwantedDomStorage();
    }
  }
};
