/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * UI managment (add toolbar button, an menu items)
 *
 * Partially based on:
 *  - https://github.com/jvillalobos/Australis-View-Demo/blob/master/src/bootstrap.js
 *  - https://blog.mozilla.org/addons/2014/03/06/australis-for-add-on-developers-2/
 * ************************************************************************
 */

"use strict";

const CK_TOOLBAR_CSS  = "chrome://cookiekeeper_app/skin/toolbar.css";

function ckUnloader() {
  // Check if the last browser window is closed
  // Note: The Services.wm.getEnumerator() function can't be called within the
  // event onCloseWindow(). I got an infinite loop.
  // So use the "unload" event instead.

  this.removeEventListener("unload", ckUnloader, false);
  let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);

  if (!enumerator.hasMoreElements()) {
    // If so, close all cookiekeeper windows too.
    CookieKeeper.closeCookieWindow();
  }
}

function isPrivateBrowsingMode(window) {
  try {
    // Firefox 20+
    Cu.import("resource://gre/modules/PrivateBrowsingUtils.jsm");
    if (typeof(PrivateBrowsingUtils.isContentWindowPrivate) == "function") {
      // for e10s
      return PrivateBrowsingUtils.isContentWindowPrivate(window);
    } else {
      // deprecated for e10s
      return PrivateBrowsingUtils.isWindowPrivate(window);
    }
  } catch (ex) {
    return false;
  }
}

/**
 * Australis UI
 */
var ckAustralisWindowListener = {

  _showMenuPanel: function(aEvent) {
    // log.d("_showMenuPanel ", aEvent);
  },

  _hideMenuPanel: function(aEvent) {
    // log.d("_hideMenuPanel ", aEvent);
  },

  insertButton: function() {
    // loadOverlay Observer callback
    // Create the button only after the overlay menu.australis.xul is loaded.
    // If the button is created before the overlay is loaded, CustomizableUI will issue an error
    // Even if it does not seems to be a fatal error (everythings still work), I think it is better that way.

    this._buttonCreated = true;
    let ckLabel         = tr("cookiekeeper.label");
    let ckTooltip       = tr("cookiekeeper.tooltip");

    let buttonCK = {
      id            : CK_BUTTON_ID,
      defaultArea   : CustomizableUI.AREA_NAVBAR,
      label         : ckLabel,
      tooltiptext   : ckTooltip,
      type          : "view",
      viewId        : CK_MENU_PANEL_ID,
      onViewShowing : ckAustralisWindowListener._showMenuPanel,
      onViewHiding  : ckAustralisWindowListener._hideMenuPanel,
      showInPrivateBrowsing : false // this add-on is not relevant in private browsing mode
    };

    CustomizableUI.createWidget(buttonCK);
  },

  _buttonCreated: false,
  insertButtonDelayed: function(subject, topic, data) {
    // After the overlay is loaded, we need a timeout to add the CK button.
    // or the CK_MENU_PANEL_ID will not be accessible.
    // This is only a problem when first installing CookieKeeper,
    // on the currently opened windows. Opening a new window, it works, or
    // if the browser start, there is no problem either.
    if (topic == "xul-overlay-merged" && !ckAustralisWindowListener._buttonCreated) {
      Components.utils.import("resource://gre/modules/Timer.jsm");
      ckAustralisWindowListener._buttonCreated = true;
      setTimeout(function() {
        ckAustralisWindowListener.insertButton();
      }, 100);
    }
  },

  addUI: function(domWindow) {
    let doc = domWindow.document;
    domWindow.addEventListener("unload", ckUnloader, false);
    doc.loadOverlay("chrome://cookiekeeper/content/overlay.menu.australis.xul", ckAustralisWindowListener.insertButtonDelayed);

    // try {
    //   let domWinUtils = domWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindowUtils);
    //   domWinUtils.loadSheet(ckUi._uriCSS, domWinUtils.USER_SHEET);
    // } catch (ex) {
    //   log.e("Error loading " + CK_TOOLBAR_CSS, ex);
    // }
  },

  removeUI: function(domWindow) {
    // Removes all added UI elements.
    let doc   = domWindow.document;
    let panel = doc.getElementById(CK_MENU_PANEL_ID);

    if (panel !== null) {
      panel.parentNode.removeChild(panel);
    }

    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    let ckMenu    = doc.getElementById(CK_MENU_ID);
    if (toolsMenu !== null && ckMenu !== null) {
      toolsMenu.removeChild(ckMenu);
    }

    // try {
    //   let domWinUtils = domWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindowUtils);
    //   domWinUtils.removeSheet(ckUi._uriCSS, domWinUtils.USER_SHEET);
    // } catch (ex) {
    //   log.e("Error remove " + CK_TOOLBAR_CSS, ex);
    // }
  },

  onOpenWindow: function(aXULWindow) {
    // A new window has opened.
    let that = this;
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);

    // Wait for it to finish loading
    domWindow.addEventListener("DOMContentLoaded", function ckListener() {
        domWindow.removeEventListener("DOMContentLoaded", ckListener, false);

        if (!isPrivateBrowsingMode(domWindow)) {
          // If this is a browser window then setup its UI
          let docElem = domWindow.document.documentElement;
          if (docElem.getAttribute("windowtype") === APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
            that.addUI(domWindow);
          }
        }
      }, false);
  },

  onCloseWindow: function(aXULWindow) {
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
    let docElem = domWindow.document.documentElement;
    if (docElem.getAttribute("windowtype") === APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
      this.removeUI(domWindow);
    }
  },

  onWindowTitleChange: function(aXULWindow, aNewTitle) {
  }
};

/**
 * Legacy UI
 */
var ckLegacyWindowListener = {

  insertButton: function(aDocument, aButton, insertAndPersist) {
    let toolbox = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBOX);
    if (toolbox.palette == null) {
      // When a new window is opened, the toolbox do not yet exists, items need to be
      // placed in the BrowserToolbarPalette like for an overlay toolbar button.
      // If placed in the BrowserToolbarPalette, Mozilla will restore itself the button position.
      toolbox = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].PALETTE);
      toolbox.appendChild(aButton);
    } else {
      // The window was already opened and loaded, we need to know where to place the button
      // Put it for now in the palette
      toolbox.palette.appendChild(aButton);
      let toolbar = null;
      let nextItem = null; // to replace the button at the right position

      // Search the toolbar which have the attribute "currentset"  having our button id.
      // rem: not using toolbox children, because some toolbar may not be a direct child
      // of the toolbox, like the add-on bar. See browser.xul in Firefox source code.
      let toolbars = aDocument.querySelectorAll("toolbar");
      for (let i=0; i<toolbars.length && toolbar === null; i++) {
        toolbar = toolbars[i];
        let currentSet      = toolbar.getAttribute("currentset");
        let currentSetArray = currentSet == null ? [] : currentSet.split(",");
        let idx = currentSetArray.indexOf(aButton.id);
        if (idx == -1) {
          toolbar = null;
        } else {
          // Found the toolbar, now get the button position
          // Do a while loop, if the next element do not exists
          // (For example, a disabled addon)
          idx++;
          let nextItemId = currentSetArray[idx];
          while (idx < currentSetArray.length && nextItem == null) {
            nextItem = aDocument.getElementById(currentSetArray[idx]);
            if (nextItem == null) {
              // Maybe a spring, a separator or a space?
              let nChildren = toolbar.childNodes.length;
              nextItem = nChildren > idx-1 ? toolbar.childNodes[idx-1] : null;
              if (nextItem != null && nextItem.id.indexOf(nextItemId, 0) == -1) {
                // Apparently not
                nextItem = null;
              }
            }
            idx++;
          }
        }
      }

      if (toolbar == null) {
        if (insertAndPersist) {
          // Not found, use the default toolbar (nav-bar for FFx)
          toolbar = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBAR);
          toolbar.insertItem(aButton.id);
          // Save to the localstore.rdf
          toolbar.setAttribute("currentset", toolbar.currentSet);
          aDocument.persist(toolbar.id, "currentset");
        } // else { log.d("do not insert button"); }
      } else {
        toolbar.insertItem(aButton.id, nextItem);
      }
    }
  },

  addUI: function(domWindow, startupReason) {
    const CK_ICON16     = "chrome://cookiekeeper/skin/cookie16.png";
    const PREF_ICON16   = "chrome://cookiekeeper_os/skin/options16.png";
    const DEL_ICON16    = "chrome://cookiekeeper_os/skin/delete16.png";

    var buttonCK        = null;
    var menuCK          = null;
    var ckLabel         = null;
    var ckTooltip       = null;
    var ckMenuShowMain  = null;
    var ckMenuRemAll    = null;
    var ckMenuShowPref  = null;

    if (typeof(domWindow) === "undefined" || domWindow === null) {
      log.e("Error window not defined");
      return;
    }

    let doc = domWindow.document;

    function openCookieWindow() {
      CookieKeeper.showMainWindow(domWindow);
    }

    function openPreferencesWindow() {
      CookieKeeper.showPreferencesDialogue(domWindow, null);
    }

    function clearCookies() {
      CookieKeeper.clearAllCookies(false);
    }

    function createMenu(doc) {
      let miShowCK = doc.createElementNS(XUL_NS, "menuitem");
      miShowCK.setAttribute("label", ckMenuShowMain);
      miShowCK.setAttribute("image", CK_ICON16);
      miShowCK.addEventListener("command", openCookieWindow, false);

      let miShowPref = doc.createElementNS(XUL_NS, "menuitem");
      miShowPref.setAttribute("label", ckMenuShowPref);
      miShowPref.setAttribute("image", PREF_ICON16);
      miShowPref.addEventListener("command", openPreferencesWindow, false);

      let miClear = doc.createElementNS(XUL_NS, "menuitem");
      miClear.setAttribute("label",  ckMenuRemAll);
      miClear.setAttribute("image", DEL_ICON16);
      miClear.addEventListener("command", clearCookies, false);

      let mp = doc.createElementNS(XUL_NS, "menupopup");
      mp.appendChild(miShowCK);
      mp.appendChild(miShowPref);
      mp.appendChild(miClear);

      return mp;
    }

    ckLabel        = tr("cookiekeeper.label");
    ckTooltip      = tr("cookiekeeper.tooltip");
    ckMenuShowMain = tr("menu.showcookies.label");
    ckMenuRemAll   = tr("menu.removecookies.label");
    ckMenuShowPref = tr("menu.preferences.label");

    // Add toolbar button
    buttonCK = doc.createElementNS(XUL_NS, "toolbarbutton");
    buttonCK.setAttribute("id", CK_BUTTON_ID);
    buttonCK.setAttribute("label", ckLabel);
    buttonCK.setAttribute("class", "toolbarbutton-1 chromeclass-toolbar-additional");
    buttonCK.setAttribute("tooltiptext", ckTooltip);
    buttonCK.setAttribute("persist", "class");
    buttonCK.setAttribute("type", "menu");

    buttonCK.appendChild(createMenu(doc));

    let insertAndPersist = (currReason == ADDON_INSTALL); // currReason is set in bootstrap.js
    this.insertButton(doc, buttonCK, insertAndPersist);

    // Add menu item under the tools-menu
    menuCK = doc.createElementNS(XUL_NS, "menu");
    menuCK.setAttribute("id", CK_MENU_ID);
    menuCK.setAttribute("class", "menu-iconic");
    menuCK.setAttribute("image", CK_ICON16);
    menuCK.setAttribute("label", ckLabel);
    menuCK.setAttribute("tooltiptext", ckTooltip);
    menuCK.appendChild(createMenu(doc));

    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    toolsMenu.appendChild(menuCK);

    domWindow.addEventListener("unload", ckUnloader, false);
  },

  removeUI: function(domWindow) {
    let doc       = domWindow.document;
    let ckButton  = doc.getElementById(CK_BUTTON_ID);
    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    let ckMenu    = doc.getElementById(CK_MENU_ID);

    if (ckButton != null) {
      ckButton.parentNode.removeChild(ckButton);
    } else {
      // The button may be in the toolbox palette
      let palette = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBOX).palette;
      let elems   = palette.childNodes;
      for (let i=0; i<elems.length; i++) {
        // Do a try catch if some elements are not accessible, like disabled addon which do not
        // have removed their button from the palette
        try {
          let elem = elems[i];
          if (CK_BUTTON_ID == elem.id) {
            palette.removeChild(elem);
          }
        } catch (ex) {
          log.e(ex);
        }
      }
    }
    if (toolsMenu !== null && ckMenu !== null) {
      toolsMenu.removeChild(ckMenu);
    } else {
      log.w("tools menu item not found");
    }
  },

  onOpenWindow: function(aXULWindow) {
    // A new window has opened.
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);

    // Wait for it to finish loading
    domWindow.addEventListener("DOMContentLoaded", function ckListener() {
        domWindow.removeEventListener("DOMContentLoaded", ckListener, false);
        // If not in private browsing mode
        if (!isPrivateBrowsingMode(domWindow)) {
          // If this is a browser window then setup its UI
          let docElem = domWindow.document.documentElement;
          if (docElem.getAttribute("windowtype") == APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
            ckLegacyWindowListener.addUI(domWindow);
          }
        }
      }, false);
  },

  onCloseWindow: function(aXULWindow) {
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
    let docElem   = domWindow.document.documentElement;
    if (docElem.getAttribute("windowtype") == APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
      this.removeUI(domWindow);
    }
  },

  onWindowTitleChange: function(aXULWindow, aNewTitle) {
  }
};

var ckUi = {
  _useAustralis : false,
  _uriCSS       : null,
  _cssType      : null,

  init: function() {

    // Inject the CSS using the stylesheet service, for the toolbar icon:
    let ssService = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    this._uriCSS  = Services.io.newURI(CK_TOOLBAR_CSS, null, null);

    let windowListener = null;
    try {
      // Setup the toolbar icon for Australis UI (Firefox 29+)
      // https://blog.mozilla.org/addons/2014/03/06/australis-for-add-on-developers-2/
      // Will fail (throw an exeption) if not found:
      Cu.import("resource:///modules/CustomizableUI.jsm");

      windowListener     = ckAustralisWindowListener;
      this._cssType      = ssService.USER_SHEET;
      this._useAustralis = true;
    } catch (ex) {
      windowListener     = ckLegacyWindowListener;
      this._cssType      = ssService.AUTHOR_SHEET;
      this._useAustralis = false;
    }

    ssService.loadAndRegisterSheet(this._uriCSS, this._cssType);

    // Add the menu panel to existing windows:
    let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);
    while (enumerator.hasMoreElements()) {
      windowListener.addUI(enumerator.getNext());
    }

    // Window listener to add the menu panel on new windows:
    Services.wm.addListener(windowListener);
  }, // init

  uninit: function() {
    try {
      CookieKeeper.closeCookieWindow()
    } catch (ex) {
      log.e(ex);
    }

    let windowListener = null;
    if (this._useAustralis) {
      windowListener = ckAustralisWindowListener;
    } else {
      windowListener = ckLegacyWindowListener;
    }

    let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);

    Services.wm.removeListener(windowListener);

    while (enumerator.hasMoreElements()) {
      let aWindow = enumerator.getNext();
      aWindow.removeEventListener("unload", ckUnloader, false);
      windowListener.removeUI(aWindow);
    }

    if (this._useAustralis) {
      CustomizableUI.destroyWidget(CK_BUTTON_ID);
    }

    let ssService = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    if (ssService.sheetRegistered(this._uriCSS, this._cssType)) {
      ssService.unregisterSheet(this._uriCSS, this._cssType);
    }

    this._cssType = null;
    this._uriCSS  = null;
  } // uninit

}; // ckUi
